from typing import Union, Optional

from musicscore import Part, Chord
from musicscore.chord import Rest
from musicscore.exceptions import (
    AlreadyFinalizedError,
    ScoreHasNoPartsError,
    ScoreMultiMeasureRestError,
    ScorePartIdIsNotUniqueError,
)
from musicscore.finalize import FinalizeMixin
from musicscore.layout import Scaling, PageLayout, SystemLayout, StaffLayout
from musicscore.musictree import MusicTree
from musicscore.quantize import QuantizeMixin
from musicscore.quarterduration import QuarterDuration
from musicscore.xmlwrapper import XMLWrapper
from musicxml.xmlelement.xmlelement import (
    XMLScorePartwise,
    XMLPartList,
    XMLCredit,
    XMLCreditWords,
    XMLIdentification,
    XMLEncoding,
    XMLSupports,
    XMLScorePart,
    XMLPartGroup,
    XMLGroupSymbol,
    XMLGroupBarline,
    XMLGroupName,
    XMLGroupAbbreviation,
    XMLMeasureStyle,
)

__all__ = ["TITLE", "SUBTITLE", "POSSIBLE_SUBDIVISIONS", "Score"]
#:
TITLE = {
    "font_size": 24,
    "default_x": {"A4": {"portrait": 616}},
    "default_y": {"A4": {"portrait": 1573}},
    "justify": "center",
    "valign": "top",
}

#:
SUBTITLE = {
    "font_size": 18,
    "default_x": {"A4": {"portrait": 616}},
    "default_y": {"A4": {"portrait": 1508}},
    "halign": "center",
    "valign": "top",
}
#:
POSSIBLE_SUBDIVISIONS = {
    QuarterDuration(1, 4): [2, 3],
    QuarterDuration(1, 2): [2, 3, 4, 5],
    QuarterDuration(1): [2, 3, 4, 5, 6, 7, 8],
}


class Score(MusicTree, QuantizeMixin, FinalizeMixin, XMLWrapper):
    """
    Parent type: ``None``

    Child type: :obj:`~musicscore.part.Part`
    """

    _ATTRIBUTES = {
        "version",
        "title",
        "subtitle",
        "scaling",
        "page_layout",
        "system_layout",
        "staff_layout",
        "new_system",
    }
    _ATTRIBUTES = _ATTRIBUTES.union(MusicTree._ATTRIBUTES)
    _ATTRIBUTES = _ATTRIBUTES.union(QuantizeMixin._ATTRIBUTES)

    XMLClass = XMLScorePartwise

    def __init__(
        self,
        version="4.0",
        title=None,
        subtitle=None,
        get_quantized=False,
        new_system=False,
        *args,
        **kwargs,
    ):
        super().__init__(get_quantized=get_quantized)
        self._xml_object = self.XMLClass(*args, **kwargs)
        self._update_xml_object()
        self._version = None
        self._title = None
        self._subtitle = None
        self._page_layout = None
        self._system_layout = None
        self._staff_layout = None
        self._scaling = None
        self._new_system = None

        self.scaling = Scaling()
        self.page_layout = PageLayout()
        self.system_layout = SystemLayout()
        self.version = version
        self.title = title
        self.subtitle = subtitle
        self.new_system = new_system
        self._possible_subdivisions = POSSIBLE_SUBDIVISIONS.copy()

        self._measure_numbers_within_multi_measure_rests = set()

        self._final_updated = False

    def _check_parts(self):
        part_ids = [part.id_ for part in self.get_children()]
        if not part_ids:
            raise ScoreHasNoPartsError

        if len(set(part_ids)) != len(part_ids):
            raise ScorePartIdIsNotUniqueError

    def _create_missing_measures(self):
        number_of_measures = max([len(p.get_children()) for p in self.get_children()])
        longest_parts = [
            p
            for p in self.get_children()
            if len(p.get_children()) == number_of_measures
        ]
        for p in set(self.get_children()).difference(set(longest_parts)):
            for measure_number in range(
                len(p.get_children()) + 1, number_of_measures + 1
            ):
                p.add_chord(
                    Rest(longest_parts[0].get_measure(measure_number).quarter_duration)
                )

    def _get_title_attributes(self):
        output = TITLE.copy()
        output["default_x"] = TITLE["default_x"]["A4"]["portrait"]
        output["default_y"] = TITLE["default_y"]["A4"]["portrait"]
        return output

    def _get_subtitle_attributes(self):
        output = SUBTITLE.copy()
        output["default_x"] = SUBTITLE["default_x"]["A4"]["portrait"]
        output["default_y"] = SUBTITLE["default_y"]["A4"]["portrait"]
        return output

    def _set_last_barline(self):
        last_measures = [
            p.get_children()[-1] for p in self.get_children() if p.get_children()
        ]
        try:
            if not last_measures[0].get_barline():
                for m in last_measures:
                    m.set_barline(style="light-heavy")
        except IndexError:
            pass

    def _set_missing_barlines(self):
        for measures in zip(*[p.get_children() for p in self.get_children()]):
            right_barline = None
            left_barline = None
            for m in measures:
                if m.get_barline(location="right"):
                    right_barline = m.get_barline(location="right")
                    break
            for m in measures:
                if m.get_barline(location="left"):
                    left_barline = m.get_barline(location="left")
                    break
            if right_barline:
                for m in measures:
                    m._barlines["right"] = right_barline
            if left_barline:
                for m in measures:
                    m._barlines["left"] = left_barline

    def _update_xml_object(self):
        self.xml_object.xml_part_list = XMLPartList()
        self.xml_object.xml_identification = XMLIdentification()
        encoding = self.xml_object.xml_identification.xml_encoding = XMLEncoding()
        encoding.add_child(XMLSupports(element="accidental", type="yes"))
        encoding.add_child(XMLSupports(element="beam", type="yes"))
        encoding.add_child(XMLSupports(element="stem", type="yes"))

    @property
    def new_system(self) -> bool:
        """
        Set or get new_system property. It will be automatically set to ``True`` if a :obj:`~musicscore.measure.Measure` sets its :obj:`musicscore.measure.Measure.new_system` to ``True``. ``<encoding>`` in :obj:`~musicxml.xmlelement.xmlelement.XMLScorePartwise`\'s ``<identification>`` will have a ``<support>`` child with attribute ``new-system`` if :obj:`musicscore.score.Score.new_system` is set to ``True``.
        """
        return self._new_system

    @new_system.setter
    def new_system(self, val):
        if isinstance(val, bool):
            self._new_system = val
        else:
            raise TypeError(
                f"new_system {val} must be of type bool and not {val.__class__}"
            )
        encoding = self.xml_object.xml_identification.xml_encoding
        new_system_supports = [
            sup
            for sup in encoding.get_children_of_type(XMLSupports)
            if sup.attribute == "new-system"
        ]
        if self.new_system:
            if not new_system_supports:
                self.xml_object.xml_identification.xml_encoding.add_child(
                    XMLSupports(
                        attribute="new-system", element="print", type="yes", value="yes"
                    )
                )
        else:
            if new_system_supports:
                self.xml_object.xml_identification.xml_encoding.remove(
                    new_system_supports[0]
                )

    @property
    def page_layout(self) -> PageLayout:
        """
        Set and get page layout. After setting value, page layout's parent is set to self.

        :type: :obj:`~musicscore.pagelayout.PageLayout`
        :return: :obj:`~musicscore.pagelayout.PageLayout`
        """
        return self._page_layout

    @page_layout.setter
    def page_layout(self, val):
        if not isinstance(val, PageLayout):
            raise TypeError
        self._page_layout = val
        self.page_layout.parent = self

    @property
    def scaling(self) -> Scaling:
        """
        :type: :obj:`~musicscore.scaling.Scaling`
        :return: :obj:`~musicscore.scaling.Scaling`
        """
        return self._scaling

    @scaling.setter
    def scaling(self, val):
        if not isinstance(val, Scaling):
            raise TypeError
        val.score = self
        self._scaling = val

    @property
    def staff_layout(self):
        """
        Set and get staff layout. After setting value, staff layout's parent is set to self.

        :type: :obj:`~musicscore.stafflayout.StaffLayout`
        :return: :obj:`~musicscore.stafflayout.StaffLayout`
        """

        return self._staff_layout

    @staff_layout.setter
    def staff_layout(self, val):
        if not isinstance(val, StaffLayout):
            raise TypeError
        self._staff_layout = val
        self.staff_layout.parent = self

    @property
    def subtitle(self):
        """
        - If val is ``None`` and a subtitle object as credit already exists, this object will be removed.
        - If val is not ``None``:
          - if a subtitle exists its ``value_`` will be replaced.
          - else a subtitle as credit will be added.

        :type: Optional[str]
        :return: Optional[str]
        """
        return self._subtitle

    @subtitle.setter
    def subtitle(self, val):
        if val is not None:
            if self._subtitle is None:
                credit = self.xml_object.add_child(XMLCredit(page=1))
                credit.xml_credit_type = "subtitle"
                self._subtitle = credit.add_child(
                    XMLCreditWords(value_=val, **self._get_subtitle_attributes())
                )
            else:
                self._subtitle.value_ = val
        else:
            if self._subtitle is None:
                pass
            else:
                credit = self._subtitle.up
                credit.up.remove(credit)
                self._subtitle = None

    @property
    def system_layout(self):
        """
        Set and get system layout. After setting value, system layout's parent is set to self.

        :type: :obj:`~musicscore.systemlayout.SystemLayout`
        :return: :obj:`~musicscore.systemlayout.SystemLayout`
        """
        return self._system_layout

    @system_layout.setter
    def system_layout(self, val):
        if not isinstance(val, SystemLayout):
            raise TypeError
        self._system_layout = val
        self.system_layout.parent = self

    @property
    def title(self):
        """
        - If val is ``None`` and a title object as credit already exists, this object will be removed.
        - If val is not ``None``:
          - if a title exists its ``value_`` will be replaced.
          - else a title as credit will be added.

        :type: Optional[str]
        :return: Optional[str]
        """
        return self._title

    @title.setter
    def title(self, val):
        if val is not None:
            if self._title is None:
                credit = self.xml_object.add_child(XMLCredit(page=1))
                credit.xml_credit_type = "title"
                self._title = credit.add_child(
                    XMLCreditWords(value_=val, **self._get_title_attributes())
                )
            else:
                self._title.value_ = val
        else:
            if self._title is None:
                pass
            else:
                credit = self._title.up
                credit.up.remove(credit)
                self._title = None

    @property
    def version(self) -> str:
        """
        :type: Any
        :return: xml_object's version
        :rtype: str
        """
        return self._version

    @version.setter
    def version(self, val):
        self._version = str(val)
        self.xml_object.version = self.version

    def add_child(self, child: "Part") -> "Part":
        """
        - Check and add child to list of children. Child's parent is set to self.
        - Part's ``xml_object`` is add to score's ``xml_object`` as child
        - Part's ``score_part.xml_object`` is added to score's ``xml_part_list``

        :param child: :obj:`~musicscore.part.Part`, required
        :return: child
        :rtype: :obj:`~musicscore.part.Part`
        """
        if self._finalized is True:
            raise AlreadyFinalizedError(self, "add_child")
        super().add_child(child)
        self.xml_object.add_child(child.xml_object)
        if not self.xml_part_list:
            self.xml_part_list.xml_score_part = child.score_part.xml_object
        else:
            self.xml_part_list.add_child(child.score_part.xml_object)
        return child

    def add_part(self, id: str) -> "Part":
        """
        Creates and adds part

        :param id: part's id
        :return: part
        """
        if self._finalized is True:
            raise AlreadyFinalizedError(self, "add_part")
        p = Part(id)
        return self.add_child(p)

    def export_xml(self, path: "pathlib.Path") -> None:
        """
        Creates a musicxml file

        :param path: Output xml file
        :return: None
        """
        with open(path, "+w") as f:
            f.write("""<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<!DOCTYPE score-partwise PUBLIC
    "-//Recordare//DTD MusicXML 4.0 Partwise//EN"
    "http://www.musicxml.org/dtds/partwise.dtd">
""")
            f.write(self.to_string())

    def finalize(self) -> None:
        self._check_parts()
        self._create_missing_measures()
        self._set_missing_barlines()
        self._set_last_barline()
        super().finalize()
        for measure_number in self._measure_numbers_within_multi_measure_rests:
            for part in self.get_children():
                measure = part.get_measure(measure_number)
                for ch in measure.get_chords():
                    ch.notes[0].xml_rest.measure = "yes"

    def group_parts(
        self,
        number: Union[int, str],
        start_part_number: int,
        end_part_number: int,
        symbol: str = "square",
        name: Optional[str] = None,
        abbreviation: Optional[str] = None,
        **kwargs,
    ) -> None:
        """
        Adds a XMLPartList child

        :param number: sets number property of :obj:`~musicxml.xmlelement.xmlelement.XMLPartGroup`
        :param start_part_number: number of part to start the group
        :param end_part_number: number of part to end the group
        :param symbol: symbol property of :obj:`~musicxml.xmlelement.xmlelement.XMLGroupSymbol`:'none', 'brace', 'line', 'bracket', 'square'; default: 'square'
        :param name: value of :obj:`~musicxml.xmlelement.xmlelement.XMLGroupName`
        :param abbreviation: value of :obj:`~musicxml.xmlelement.xmlelement.XMLGroupAbbreviation`
        :param kwargs: kwargs of :obj:`~musicxml.xmlelement.xmlelement.XMLGroupSymbol`
        :return: None
        """
        parts = self.get_children_of_type(Part)
        for part_number in [start_part_number, end_part_number]:
            if not 0 < part_number <= len(parts):
                raise ValueError(
                    f"Wrong part number {part_number}. Permitted between 1 and {len(parts)}"
                )
        if not start_part_number < end_part_number:
            raise ValueError(
                f"Wrong part numbers. start_part_number {start_part_number} must be smaller than end_part_number {end_part_number}"
            )

        new_xml_part_list = XMLPartList(xsd_check=False)
        for child in self.xml_part_list.get_children():
            if (
                isinstance(child, XMLScorePart)
                and child.id == parts[start_part_number - 1].id_
            ):
                pg = new_xml_part_list.add_child(
                    XMLPartGroup(number=str(number), type="start")
                )
                pg.add_child(XMLGroupSymbol(symbol, **kwargs))
                pg.add_child(XMLGroupBarline("yes"))
                if name:
                    pg.add_child(XMLGroupName(name))
                if abbreviation:
                    pg.add_child(XMLGroupAbbreviation(abbreviation))
            new_xml_part_list.add_child(child)
            if (
                isinstance(child, XMLScorePart)
                and child.id == parts[end_part_number - 1].id_
            ):
                new_xml_part_list.add_child(
                    XMLPartGroup(number=str(number), type="stop")
                )

        self.xml_part_list = new_xml_part_list

    def set_multi_measure_rest(
        self, first_measure_number: int, last_measure_number: int
    ) -> None:
        """
        Creates a multi measure rest

        :param first_measure_number: number of measure to start the multi measure rest
        :param last_measure_number: number of measure to end the multi measure rest
        :return: None
        """

        if len(self.get_children()) == 0:
            raise ScoreMultiMeasureRestError("score has no parts.")
        if last_measure_number <= first_measure_number:
            raise ScoreMultiMeasureRestError(
                f"last_measure_number {last_measure_number} must be larger than first_measure_number {first_measure_number}"
            )
        for x in range(first_measure_number, last_measure_number + 1):
            if x in self._measure_numbers_within_multi_measure_rests:
                raise ScoreMultiMeasureRestError(
                    f"measure number {x} is already part of a multiple measure reset"
                )
        for part in self.get_children():
            for _ in range(last_measure_number - len(part.get_children())):
                m = part.add_measure()
                part.add_chord(Chord(0, m.quarter_duration))

            for measure in part.get_children()[
                first_measure_number - 1 : last_measure_number
            ]:
                for ch in measure.get_chords():
                    if not ch.is_rest:
                        raise ScoreMultiMeasureRestError(
                            "Measures contain not rest chords"
                        )

            first_measure = part.get_measure(first_measure_number)
            if not first_measure.xml_attributes.xml_measure_style:
                first_measure.xml_attributes.xml_measure_style = XMLMeasureStyle()
            first_measure.xml_attributes.xml_measure_style.xml_multiple_rest = (
                last_measure_number - first_measure_number + 1
            )
            if part == self.get_children()[0]:
                self._measure_numbers_within_multi_measure_rests.update(
                    {x for x in range(first_measure_number, last_measure_number + 1)}
                )

    def write(self, *args, **kwargs):
        """
        Not implemented. Use Score.export_xml instead!
        """
        raise NotImplementedError("Use Score.export_xml instead!")



from fractions import Fraction
from unittest import TestCase

from musicscore import Part, Chord
from musicscore.time import Time, flatten_times, _convert_signatures_to_ints


class TestTime(TestCase):
    def test_convert_signatures_to_ints(self):
        assert _convert_signatures_to_ints([1, 2]) == [1, 2]
        assert _convert_signatures_to_ints([1, 2, 3, 4]) == [1, 2, 3, 4]
        assert _convert_signatures_to_ints(["1", "2", "3", "4"]) == [1, 2, 3, 4]
        assert _convert_signatures_to_ints(["2+3+5", 8, 3, 4]) == [
            2,
            8,
            3,
            8,
            5,
            8,
            3,
            4,
        ]
        assert _convert_signatures_to_ints(["2+3", "8"]) == [2, 8, 3, 8]

    def test_time_init(self):
        t = Time()
        expected = """<time>
  <beats>4</beats>
  <beat-type>4</beat-type>
</time>
"""
        assert t.to_string() == expected
        t.xml_beats = "3"
        expected = """<time>
  <beats>3</beats>
  <beat-type>4</beat-type>
</time>
"""
        assert t.to_string() == expected
        t = Time(6, 8, 3, 4)
        expected = """<time>
  <beats>6</beats>
  <beat-type>8</beat-type>
  <beats>3</beats>
  <beat-type>4</beat-type>
</time>
"""
        assert t.to_string() == expected

    def test_change_fractions(self):
        t = Time()
        t.signatures = [7, 8, 3, 4]
        expected = """<time>
  <beats>7</beats>
  <beat-type>8</beat-type>
  <beats>3</beats>
  <beat-type>4</beat-type>
</time>
"""
        assert t.to_string() == expected
        t.signatures = None
        expected = """<time>
  <beats>4</beats>
  <beat-type>4</beat-type>
</time>
"""
        assert t.to_string() == expected
        t.signatures = [2, 4]
        expected = """<time>
  <beats>2</beats>
  <beat-type>4</beat-type>
</time>
"""
        assert t.to_string() == expected

        t.signatures = ["3+2", 8]
        expected = """<time>
  <beats>3+2</beats>
  <beat-type>8</beat-type>
</time>
"""
        assert t.to_string() == expected

    def test_time_actual_signatures(self):
        t = Time()
        assert t.actual_signatures == [1, 4, 1, 4, 1, 4, 1, 4]
        t.signatures = [3, 4]
        assert t.actual_signatures == [1, 4, 1, 4, 1, 4]
        t.actual_signatures = [3, 4]
        assert t.actual_signatures == [3, 4]

        t.signatures = [2, 8]
        assert t.signatures == [2, 8]
        assert t.actual_signatures == [2, 8]

        t.signatures = [3, 8]
        assert t.actual_signatures == [3, 8]

        t.signatures = [4, 8]
        assert t.actual_signatures == [2, 8, 2, 8]

        t.signatures = [5, 8]
        assert t.actual_signatures == [3, 8, 2, 8]

        t.signatures = [6, 8]
        assert t.actual_signatures == [3, 8, 3, 8]

        t.signatures = [7, 8]
        assert t.actual_signatures == [4, 8, 3, 8]

        t.signatures = [8, 8]
        assert t.actual_signatures == 8 * [1, 8]

        t.signatures = [9, 8]
        assert t.actual_signatures == 3 * [3, 8]

        t.signatures = [10, 8]
        assert t.actual_signatures == 10 * [1, 8]

        t.signatures = [11, 8]
        assert t.actual_signatures == 11 * [1, 8]

        t.signatures = [12, 8]
        assert t.actual_signatures == 4 * [3, 8]

        t.signatures = [15, 8]
        assert t.actual_signatures == 5 * [3, 8]

        t.signatures = [18, 8]
        assert t.actual_signatures == 6 * [3, 8]

        t.actual_signatures = ["3+2+5", "8"]
        assert t.actual_signatures == [3, 8, 2, 8, 5, 8]

        t.signatures = [3, 4]
        assert t.actual_signatures == [1, 4, 1, 4, 1, 4]
        t.actual_signatures = [3, 4]
        assert t.actual_signatures == [3, 4]
        t.signatures = [3, 2]
        assert t.actual_signatures == [1, 2, 1, 2, 1, 2]

    def test_get_beats_quarter_durations(self):
        t = Time()
        assert t.get_beats_quarter_durations() == [Fraction(1)] * 4
        t.signatures = [3, 4]
        assert t.get_beats_quarter_durations() == [Fraction(1)] * 3
        t.signatures = [6, 8]
        assert t.get_beats_quarter_durations() == [Fraction(3, 2)] * 2
        t.actual_signatures = [1, 8] * 6
        assert t.get_beats_quarter_durations() == [Fraction(1, 2)] * 6
        t.signatures = [3, 4, 2, 8]
        # t._reset_actual_signatures()
        assert t.get_beats_quarter_durations() == [Fraction(1)] * 3 + [Fraction(1, 1)]

    def test_flatten_times(self):
        times = [2 * Time(3, 8), (3, 4), 3 * [(1, 8)], Time(1, 8, 3, 4), Time(3, 4)]
        flattened_times = flatten_times(times)
        assert [t.signatures for t in flattened_times] == [
            (3, 8),
            (3, 8),
            (3, 4),
            (1, 8),
            (1, 8),
            (1, 8),
            (1, 8, 3, 4),
            (3, 4),
        ]

    def test_copy(self):
        t = Time(3, 4, show=False)
        t.actual_signatures = (1, 8, 2, 4)
        copied = t.__copy__()
        assert copied != t
        assert copied.signatures == t.signatures
        assert copied.actual_signatures == t.actual_signatures
        assert copied.show == t.show


class TestActualTime(TestCase):
    # def test_actual_time_cannot_change_measure_quarter_duration(self):
    #     t = Time(3, 4)
    #     t.actual_signatures = [4, 4]
    def test_add_measure_sets_actual_time_of_previous(self):
        p = Part("p1")
        t = Time(4, 4)
        t.actual_signatures = [2, 2]
        m1 = p.add_measure(t)
        m2 = p.add_measure()
        assert m1.time.actual_signatures == m2.time.actual_signatures
        assert [b.quarter_duration for b in m1.get_beats()] == [
            b.quarter_duration for b in m2.get_beats()
        ]

    def test_actual_time_and_beat_quarter_duration(self):
        t = Time(3, 2)
        assert t.get_beats_quarter_durations() == [2, 2, 2]
        t.actual_signatures = [3, 2]
        assert t.get_beats_quarter_durations() == [6]

        t = Time(2, 2)
        assert t.get_beats_quarter_durations() == [2, 2]
        t.actual_signatures = [2, 2]
        assert t.get_beats_quarter_durations() == [4]

    def test_actual_time_and_beat_get_actual_notes(self):
        part = Part("p1")
        t = Time(2, 2)
        measure = part.add_measure(t)
        quarter_durations = [4 / 3, 2 / 3, 2 / 5, 2 / 5, 2 / 5, 2 / 5, 2 / 5]
        for qd in quarter_durations:
            part.add_chord(Chord(60, qd))
        assert (
            measure.get_beat(
                staff_number=1, voice_number=1, beat_number=1
            ).quarter_duration
            == 2
        )
        assert (
            len(
                measure.get_beat(
                    staff_number=1, voice_number=1, beat_number=1
                ).get_children()
            )
            == 2
        )
        assert (
            len(
                measure.get_beat(
                    staff_number=1, voice_number=1, beat_number=2
                ).get_children()
            )
            == 5
        )

        part = Part("p2")
        t = Time(2, 2)
        t.actual_signatures = [1, 2, 1, 2]
        measure = part.add_measure(t)
        quarter_durations = [4 / 3, 2 / 3, 2 / 5, 2 / 5, 2 / 5, 2 / 5, 2 / 5]
        for qd in quarter_durations:
            part.add_chord(Chord(60, qd))
        assert (
            measure.get_beat(
                staff_number=1, voice_number=1, beat_number=1
            ).quarter_duration
            == 2
        )
        assert (
            len(
                measure.get_beat(
                    staff_number=1, voice_number=1, beat_number=1
                ).get_children()
            )
            == 2
        )
        assert (
            len(
                measure.get_beat(
                    staff_number=1, voice_number=1, beat_number=2
                ).get_children()
            )
            == 5
        )



import cProfile
from contextlib import redirect_stdout
from pathlib import Path

from musicscore.chord import Chord
from musicscore.part import Part
from musicscore.score import Score
from musicscore.tests.util_subdivisions import (
    generate_all_quintuplets_manually,
    generate_all_sextuplets_manually,
    generate_all_septuplets_manually,
    generate_all_triplets_manually,
)


def p():
    """
    Write all possible tuplet combinations up until SEXTUPLETS
    """
    """
    Tester creates a timewise score
    """
    s = Score()
    """
    He adds a measure with one part to it
    """
    p = s.add_child(Part("P1", name="Music"))
    """
    He adds a 1/4 measure
    """
    p.add_measure(time=(1, 4))
    """
    All possible combinations are:
    """
    triplets = generate_all_triplets_manually()
    tuplets = (
        triplets
        + generate_all_quintuplets_manually()
        + generate_all_sextuplets_manually()
        + generate_all_septuplets_manually()
    )
    for tuplet_list in tuplets:
        for tuplet in tuplet_list:
            p.add_chord(Chord(60, tuplet))

    """
    ... and exports the xml
    """
    xml_path = Path(__file__).with_suffix(".xml")
    s.export_xml(xml_path)


# with open(str(Path(__file__))+'xsd_tree_only_once.txt', '+w') as f:
with open(Path(__file__).with_suffix(".txt"), "+w") as f:
    with redirect_stdout(f):
        cProfile.run("p()", sort="tottime")



"""
A trill spanner that spans a grace note and ends on an after-grace note at the end of the measure.
"""

from pathlib import Path

from musicscore import Score, SimpleFormat, E, F, G, A, B
from unittest import TestCase
from musicscore.util import trill_chords


class TestLily33a(TestCase):
    def test_lily_33a_Spanners(self):
        score = Score()
        part = score.add_part("p1")
        part.add_measure(time=(6, 8))
        simple_format = SimpleFormat(
            quarter_durations=[1, 1 / 4, 1 / 4, 1.5],
            midis=[E(5), F(5, "#"), G(5), A(5)],
        )
        chord = simple_format.chords[-1]
        chord.add_grace_chord(B(5), type="16th")
        g = chord.add_grace_chord(G(5), type="16th", position="after")
        chord.add_grace_chord(A(5), type="16th", position="after")
        trill_chords(chords=[chord, g], relative_y=15)
        for ch in simple_format.chords:
            part.add_chord(ch)

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



"""Several spanners defined in MusicXML: tuplet, slur (solid, dashed), tie, wedge (cresc, dim), tr + wavy-line,
single-note trill spanner, octave-shift (8va,15mb), bracket (solid down/down, dashed down/down, solid none/down,
dashed none/up, solid none/none), dashes, glissando (wavy), bend-alter, slide (solid), grouping, two-note tremolo,
hammer-on, pull-off, pedal (down, change, up)."""

from pathlib import Path

from musicscore import Score, Time, Chord, F, C
from unittest import TestCase
from musicscore.util import (
    slur_chords,
    wedge_chords,
    trill_chords,
    bracket_chords,
    octave_chords,
)
from musicxml import (
    XMLDashes,
    XMLGlissando,
    XMLBendAlter,
    XMLBend,
    XMLSlide,
    XMLHammerOn,
    XMLPullOff,
    XMLPedal,
)


class TestLily33a(TestCase):
    def test_lily_33a_Spanners(self):
        score = Score()
        part = score.add_part("p1")

        t = Time(3, 4)
        t.actual_signatures = [1, 2, 1, 4]
        part.add_measure(t)
        part.add_measure(Time(3, 4))

        [part.add_chord(Chord(71, 2 / 3)) for _ in range(3)]
        part.add_chord(Chord(0, 1))

        chords = [Chord(71, 1) for _ in range(3)]
        slur_chords(chords)
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1) for _ in range(3)]
        slur_chords(chords, line_type="dashed")
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1) for _ in range(3)]
        wedge_chords(chords, "crescendo", placement="above")
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1) for _ in range(3)]
        wedge_chords(chords, "diminuendo", placement="above")
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1) for _ in range(3)]
        trill_chords(chords, placement="above")
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1), Chord(0, 2)]
        trill_chords(chords, placement="above")
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1) for _ in range(3)]
        octave_chords(chords)
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1) for _ in range(3)]
        octave_chords(chords, type="up", size=15)
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1) for _ in range(3)]
        bracket_chords(chords, "solid", "down", "down", placement="above")
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1) for _ in range(3)]
        bracket_chords(chords, "dashed", "down", "down", placement="above")
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1) for _ in range(3)]
        bracket_chords(chords, "solid", "none", "down", placement="above")
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1) for _ in range(3)]
        bracket_chords(chords, "dashed", "none", "up", placement="above")
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1) for _ in range(3)]
        bracket_chords(chords, "solid", "none", "none", placement="above")
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1) for _ in range(3)]
        chords[0].add_words("rit.", placement="above", font_style="italic")
        chords[0].add_x(XMLDashes(type="start", relative_x=10))
        chords[1].add_x(XMLDashes(type="continue"))
        chords[2].add_x(XMLDashes(type="stop"))
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1), Chord(F(5), 1), Chord(0, 1)]
        chords[0].add_x(XMLGlissando(type="start", line_type="wavy"))
        chords[1].add_x(XMLGlissando(type="stop"))
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1), Chord(F(5), 1), Chord(0, 1)]
        b = chords[0].add_x(XMLBend(shape="curved"))
        b.xml_bend_alter = XMLBendAlter(5)
        b = chords[1].add_x(XMLBend())
        b.xml_bend_alter = XMLBendAlter(0)
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1), Chord(C(4), 1), Chord(0, 1)]
        chords[0].add_x(XMLSlide(type="start"))
        chords[1].add_x(XMLSlide(type="stop"))
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1) for _ in range(3)]
        chords[0].add_words("XMLGrouping not implemented yet")
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1), Chord(71, 1), Chord(0, 1)]
        chords[0].add_words("Fingered tremolo not implemented yet", placement="below")
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1), Chord(71, 1), Chord(0, 1)]
        chords[0].add_x(XMLHammerOn(type="start"))
        chords[1].add_x(XMLHammerOn(type="stop"))
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1), Chord(71, 1), Chord(0, 1)]
        chords[0].add_x(XMLPullOff(type="start"))
        chords[1].add_x(XMLPullOff(type="stop"))
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1) for _ in range(3)]
        chords[0].add_x(XMLPedal(type="start"))
        chords[1].add_x(XMLPedal(type="change"))
        chords[2].add_x(XMLPedal(type="stop"))
        [part.add_chord(ch) for ch in chords]

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



"""
Compound time signatures with same denominator: (3+2)/8 and (5+3+1)/4.
"""

from pathlib import Path

from musicscore import Score, Time, Chord, B
from unittest import TestCase


class TestLily11c(TestCase):
    def test_lily_11c_TimeSignatures_CompoundSimple(self):
        score = Score()
        part = score.add_part("part")
        part.add_measure(time=Time("3+2", 8))
        part.add_measure(time=Time("5+3+1", 4))
        [part.add_chord(Chord(B(4), x / 2)) for x in [1, 1, 1, 1, 1]]
        [part.add_chord(Chord(B(4), x)) for x in [4, 1, 3, 1]]

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



import cProfile
import timeit
from contextlib import redirect_stdout
from pathlib import Path

from musicscore import Score, Part, Chord, Id


def hello():
    s = Score()
    p = s.add_child(Part("p1"))
    p.add_chord(Chord(60, 4))
    s.to_string()
    Id.__refs__.clear()


def time_hello():
    print(timeit.timeit(hello, number=1000))


with open(Path(__file__).with_suffix(".txt"), "+w") as f:
    with redirect_stdout(f):
        cProfile.run("time_hello()", sort="tottime")



from musicscore.exceptions import LyricsWrongNumberOfChordsError
from musicscore.util import _generate_lyrics


class Lyrics:
    def __init__(self, word_groups, number=1, show_number=False, **kwargs):
        self._word_groups = word_groups
        self._number = number
        self._show_number = show_number
        self._xml_lyrics = _generate_lyrics(
            word_groups, number=number, show_number=show_number, **kwargs
        )

    @property
    def number(self):
        return self._number

    @property
    def show_number(self):
        return self._show_number

    @property
    def word_groups(self):
        return self._word_groups

    @property
    def xml_lyrics(self):
        return self._xml_lyrics

    def add_to_chords(self, chords):
        if len(chords) != len(self.xml_lyrics):
            raise LyricsWrongNumberOfChordsError(
                f"number of chords: {len(chords)}, number of xml_lyrics: {len(self.xml_lyrics)}"
            )

        for ch, xl in zip(chords, self.xml_lyrics):
            if xl:
                ch.add_lyric(xl)



from pathlib import Path
from unittest import TestCase, skip

from deepdiff import DeepDiff

from musicscore import (
    Score,
    Midi,
    QuarterDuration,
    Chord,
    SimpleFormat,
    TrebleClef,
    SimpleFormatException,
)
from musicscore.tests.util import (
    get_xml_diff_part,
    generate_xml_file,
    create_test_xml_paths,
)
import xml.etree.ElementTree as ET

path = Path(__file__)


class TestSimpleFormat(TestCase):
    def setUp(self) -> None:
        self.score = Score()
        super().setUp()

    def get_diff_xml_file_elements(self, path_1, path_2, element_tag):
        if path_2 is None:
            path_2 = path_1.split(".")[0] + "_expected.xml"
        xml_element_1 = ET.parse(path_1).getroot().find(element_tag)
        xml_element_2 = ET.parse(path_2).getroot().find(element_tag)
        return DeepDiff(xml_element_1, xml_element_2)

    def _check_distinctive_quarter_durations_and_values(self, quarter_durations, sf):
        for qd in sf.get_quarter_durations():
            assert isinstance(qd, QuarterDuration)
        for output_qt, input_qt in zip(sf.get_quarter_durations(), quarter_durations):
            if isinstance(input_qt, QuarterDuration):
                assert id(output_qt) == id(input_qt)
        assert [qd.value for qd in sf.get_quarter_durations()] == quarter_durations
        assert len(set([id(qd) for qd in sf.get_quarter_durations()])) == len(
            quarter_durations
        )

    def _check_distinctive_midi_instances_and_values(self, midis, sf):
        # sf.midis are all of type Midi
        for midi_list in sf.midis:
            for midi in midi_list:
                assert isinstance(midi, Midi)

        # make sure that midis is a list of lists (like sf.midis)
        midis = [list(x) if hasattr(x, "__iter__") else [x] for x in midis]

        # each midi object is distinctive (no repetition of midi objects)
        assert len([id(midi) for midi_list in sf.midis for midi in midi_list]) == len(
            [m for midi_list in midis for m in midi_list]
        )

        # if midi contains midi objects, these are passed on to sf.midis
        for output_midis, input_midis in zip(sf.get_midis(), midis):
            for output_midi, input_midi in zip(output_midis, input_midis):
                if isinstance(input_midi, Midi):
                    assert id(output_midi) == id(input_midi)

        # value of not midi objects in midis are the same as the value of their corresponding midi objects in sf.midis
        for m1, m2 in zip(sf.midis, midis):
            for x, z in zip(m1, m2):
                if not isinstance(z, Midi):
                    x.value = z

    def test_simple_format_init(self):
        SimpleFormat()

    def test_simple_format_init_with_numeral_parameters(self):
        quarter_durations = [1, 2, 3, 2, 1]
        midis = [60, (60, 62), (64, 66, 71), 72, 73]
        default_midi = 60

        sf = SimpleFormat(
            quarter_durations=quarter_durations, midis=midis, default_midi=default_midi
        )
        # // quarter_durations
        # distinctive QuarterDuration objects are created
        self._check_distinctive_quarter_durations_and_values(quarter_durations, sf)
        # // midis
        # distinctive Midi objects are created
        self._check_distinctive_midi_instances_and_values(midis, sf)
        # // default_midi
        assert sf.default_midi.value == default_midi

    def test_simple_format_init_with_object_parameters(self):
        quarter_durations = [
            QuarterDuration(1),
            QuarterDuration(2),
            QuarterDuration(3),
            QuarterDuration(2),
            QuarterDuration(1),
        ]
        midis = [
            Midi(60),
            (Midi(60), Midi(62)),
            (Midi(64), Midi(66), Midi(71)),
            Midi(72),
            Midi(73),
        ]
        default_midi = Midi(60)

        sf = SimpleFormat(
            quarter_durations=quarter_durations, midis=midis, default_midi=default_midi
        )
        # // quarter_durations
        # SimpleFormat takes over the same QuarterDuration objects
        assert sf.get_quarter_durations() == quarter_durations
        # // midis
        # SimpleFormat take over the same Midi objects
        assert sf.midis == [
            list(midi) if hasattr(midi, "__iter__") else [midi] for midi in midis
        ]
        # // default_midi
        # SimpleFormat take over the default_midi object
        assert sf.default_midi == default_midi

    def test_simple_format_init_with_mixed_parameters(self):
        quarter_durations = [1, 2, QuarterDuration(3), 2, QuarterDuration(1)]
        midis = [60, (Midi(60), Midi(62)), (64, 66, 71), Midi(72), 73]

        sf = SimpleFormat(quarter_durations=quarter_durations, midis=midis)
        # // quarter_durations
        # distinctive QuarterDuration objects are created or the same objects are passed on
        self._check_distinctive_quarter_durations_and_values(quarter_durations, sf)
        # // midis
        # distinctive Midi objects are created or the same objects are passed on
        self._check_distinctive_midi_instances_and_values(midis, sf)

    def test_simple_format_equalise_midi_quarter_duration_lengths(self):
        # no midis
        sf = SimpleFormat(quarter_durations=[1, 2, 3])
        assert [midi[0].value for midi in sf.midis] == [71] * 3

        # no quarter_durations
        sf = SimpleFormat(midis=[60, 61, 62])

        assert [qd.value for qd in sf.get_quarter_durations()] == [1] * 3
        # less midis
        sf = SimpleFormat(midis=[60, 61], quarter_durations=[1, 2, 3, 1])
        assert [midi[0].value for midi in sf.midis] == [60, 61, 71, 71]
        # default midis are distinctive
        assert id(sf.get_midis()[-1]) != id(sf.get_midis()[-2])
        # less quarter_durations
        sf = SimpleFormat(midis=[60, 61, 62, 60], quarter_durations=[1, 2])
        assert [qd.value for qd in sf.get_quarter_durations()] == [1, 2, 1, 1]
        # default quarter_durations are distinctive
        assert id(sf.get_quarter_durations()[-1]) != id(sf.get_quarter_durations()[-2])

    def test_simple_format_default_midi(self):
        # default value of default_midi is 71
        sf = SimpleFormat(quarter_durations=[1, 2, 3])
        assert [midi[0].value for midi in sf.midis] == [71, 71, 71]

        # default_midi can be set to a midi value
        sf = SimpleFormat(quarter_durations=[1, 2, 3], default_midi=60)
        # 3 distinctive Midi objects with correct values are created
        self._check_distinctive_midi_instances_and_values([60, 60, 60], sf)

        # default_midi can be set to a Midi object
        sf = SimpleFormat(quarter_durations=[1, 2, 3], default_midi=Midi(60))
        # three distinctive Midi objects are created
        self._check_distinctive_midi_instances_and_values([60, 60, 60], sf)

        assert sf.default_midi not in [midi[0] for midi in sf.midis]

        # // set default_midi after creation
        # default_midi can be set after creation to a midi value
        sf = SimpleFormat(quarter_durations=[1, 2, 3])
        sf.default_midi = 60
        self._check_distinctive_midi_instances_and_values([60, 60, 60], sf)
        # # default_midi can be set after creation to a Midi object
        sf = SimpleFormat(quarter_durations=[1, 2, 3])
        sf.default_midi = Midi(60)
        self._check_distinctive_midi_instances_and_values([60, 60, 60], sf)
        assert sf.default_midi not in [midi[0] for midi in sf.midis]

    def test_simple_format_chords(self):
        """
        Test midis to chords conversion
        """
        midis = [60, (60, 62), (64, 66, 71), 72, 73]
        sf = SimpleFormat(quarter_durations=[1, 1, 1, 1, 1], midis=midis)
        assert len(sf.chords) == len(midis)
        assert [[midi.value for midi in chord.midis] for chord in sf.chords] == [
            list(midi) if isinstance(midi, tuple) else [midi] for midi in midis
        ]

    def test_add_chord(self):
        sf = SimpleFormat(quarter_durations=[1], midis=[60])
        chords = [Chord(midis, 1) for midis in [[61, 62], 63]]
        for chord in chords:
            sf.add_chord(chord)
        assert sf.chords[1:] == chords

    def test_get_quarter_positions(self):
        quarter_duration_values = [1, 2, 3, 4, 5]
        sf = SimpleFormat(quarter_duration_values)
        assert sf.get_quarter_positions() == [0, 1, 3, 6, 10, 15]

    def test_get_chord_at_position(self):
        quarter_duration_values = [1, 2, 3, 4]
        sf = SimpleFormat(quarter_duration_values)
        assert sf.get_chord_at_position(0) == sf.chords[0]
        assert sf.get_chord_at_position(0.5) == sf.chords[0]
        assert sf.get_chord_at_position(0.99) == sf.chords[0]
        assert sf.get_chord_at_position(1) == sf.chords[1]
        assert sf.get_chord_at_position(2) == sf.chords[1]
        assert sf.get_chord_at_position(2.99) == sf.chords[1]
        assert sf.get_chord_at_position(3) == sf.chords[2]
        assert sf.get_chord_at_position(5) == sf.chords[2]
        assert sf.get_chord_at_position(6) == sf.chords[3]
        assert sf.get_chord_at_position(8) == sf.chords[3]
        assert sf.get_chord_at_position(9.99) == sf.chords[3]
        assert sf.get_chord_at_position(10) is None
        assert sf.get_chord_at_position(14) is None

    @skip
    def test_auto_clef(self):
        self.fail()

    def test_change_chords(self):
        def transpose_chord(chord):
            for midi in chord.midis:
                midi.value += sf.chords.index(chord)

        sf = SimpleFormat(quarter_durations=[1, 1, 1, 1, 1], default_midi=60)
        sf.change_chords(transpose_chord)
        assert [[midi.value for midi in chord.midis] for chord in sf.chords] == [
            [60],
            [61],
            [62],
            [63],
            [64],
        ]

    def test_extend(self):
        sf1 = SimpleFormat(midis=[60, 61, 62])
        sf2 = SimpleFormat(midis=[63, 64])
        sf1.extend(sf2)
        assert [[midi.value for midi in chord.midis] for chord in sf1.chords] == [
            [60],
            [61],
            [62],
            [63],
            [64],
        ]

    def test_mirror(self):
        sf = SimpleFormat(midis=[60, 61, 62, 64, 67])
        sf.mirror(pivot=63)
        assert [[midi.value for midi in chord.midis] for chord in sf.chords] == [
            [66],
            [65],
            [64],
            [62],
            [59],
        ]

    def test_multiply_quarter_durations(self):
        sf = SimpleFormat(quarter_durations=[1, 2, 3])
        sf.multiply_quarter_durations(2)
        assert [qd.value for qd in sf.get_quarter_durations()] == [2, 4, 6]

    def test_retrograde(self):
        sf = SimpleFormat(quarter_durations=[1, 2, 3], midis=[60, 61, 62])
        sf.retrograde()
        assert [qd.value for qd in sf.get_quarter_durations()] == [3, 2, 1]
        assert [[midi.value for midi in chord.midis] for chord in sf.chords] == [
            [62],
            [61],
            [60],
        ]

    def test_sum_1(self):
        xml_path, expected_path = create_test_xml_paths(path, "sum_1")
        sf1 = SimpleFormat(quarter_durations=[1, 2, 3], midis=[60, 61, 62])
        sf2 = SimpleFormat(quarter_durations=[3, 2, 1], midis=[63, 64, 65])
        sf = SimpleFormat.sum(sf1, sf2)
        generate_xml_file(self.score, sf1, sf2, sf, path=xml_path)
        get_xml_diff_part(expected_path, xml_path, Path(__file__))

    def test_sum_2(self):
        xml_path, expected_path = create_test_xml_paths(path, "sum_2")
        sf1 = SimpleFormat(quarter_durations=[1, 2, 3], midis=[60, 61, 62])
        sf2 = SimpleFormat(quarter_durations=[1, 3, 2], midis=[(60, 65), 67, (50, 54)])
        sf = SimpleFormat.sum(sf1, sf2)
        generate_xml_file(self.score, sf1, sf2, sf, path=xml_path)
        get_xml_diff_part(expected_path, xml_path, Path(__file__))

    def test_sum_3(self):
        xml_path, expected_path = create_test_xml_paths(path, "sum_3")
        sf1 = SimpleFormat(quarter_durations=[1, 2, 3, 6], midis=[60, 61, 62, 0])
        sf2 = SimpleFormat(
            quarter_durations=[1, 3, 2, 6], midis=[(60, 65), 67, (50, 54), 0]
        )
        sf3 = SimpleFormat(quarter_durations=[3, 4, 5], midis=[69, 68, 67])
        sf = SimpleFormat.sum(sf1, sf2, sf3)
        part = self.score.add_part(id="part-1")
        for index, simpleformat in enumerate([sf1, sf2, sf3, sf]):
            for chord in simpleformat.chords:
                part.add_chord(chord, staff_number=index + 1)
        part.get_staff(1, 3).clef = TrebleClef()
        part.get_staff(1, 4).clef = TrebleClef()
        self.score.export_xml(xml_path)
        get_xml_diff_part(expected_path, xml_path, Path(__file__))

    def test_complex_sum(self):
        xml_path, expected_path = create_test_xml_paths(path, "complex_sum")
        sf1 = SimpleFormat(
            quarter_durations=[1, 2, 3, 2, 1],
            midis=[60, (60, 62), (64, 66, 71), 72, 73],
        )
        sf2 = SimpleFormat(
            quarter_durations=[0.5, 1, 1.5, 2, 3, 1],
            midis=[0, 69, (72, 73), (58, 60, 65, 71), 80, 0],
        )
        sf = SimpleFormat.sum(sf1, sf2)
        generate_xml_file(self.score, sf1, sf2, sf, path=xml_path)
        get_xml_diff_part(expected_path, xml_path, Path(__file__))

    def test_sum_of_tied_chords(self):
        # path = Path(__file__).stem + '_sum_of_tied_chords.xml'
        # expected = Path(__file__).stem + '_sum_of_tied_chords_expected.xml'
        sf1 = SimpleFormat(quarter_durations=[1, 2, 3], midis=[[60, 63], [60, 61], 62])
        sf2 = SimpleFormat(quarter_durations=[3, 2, 1], midis=[63, 64, 65])
        # sum1 = SimpleFormat.sum(sf1, sf2)

        sf2.chords[0].add_tie("start")
        sf2.chords[1].add_tie("start")
        sf2.chords[2].add_tie("start")
        with self.assertRaises(SimpleFormatException):
            SimpleFormat.sum(sf1, sf2)
        # self.generate_xml_file(sf1, sf2, sum1, sum2, path=path)
        # get_xml_diff_part(expected, path)

    def test_sum_without_duplicates_1(self):
        xml_path, expected_path = create_test_xml_paths(
            path, "sum_without_duplicates_1"
        )
        sf1 = SimpleFormat(quarter_durations=[1, 2, 3], midis=[60, (61, 67), 62])
        sf2 = SimpleFormat(
            quarter_durations=[1, 3, 2], midis=[(60, 65), 67, (70, 62, 65)]
        )
        sum1 = SimpleFormat.sum(sf1, sf2)
        sum2 = SimpleFormat.sum(sf1, sf2, no_duplicates=True)
        generate_xml_file(self.score, sf1, sf2, sum1, sum2, path=xml_path)
        get_xml_diff_part(xml_path, xml_path, Path(__file__))

    def test_sum_without_duplicates_2(self):
        xml_path, expected_path = create_test_xml_paths(
            path, "sum_without_duplicates_2"
        )
        sf1 = SimpleFormat(quarter_durations=[1, 2, 3], midis=[60, (61, 66), 62])
        sf2 = SimpleFormat(
            quarter_durations=[1, 3, 2], midis=[(60, 65), 66, (70, 62, 65)]
        )
        sf1.chords[1].midis[1].accidental.mode = "flat"
        sf2.chords[1].midis[0].accidental.mode = "sharp"
        sum1 = SimpleFormat.sum(sf1, sf2)
        sum2 = SimpleFormat.sum(sf1, sf2, no_duplicates=True)
        generate_xml_file(self.score, sf1, sf2, sum1, sum2, path=xml_path)
        get_xml_diff_part(expected_path, xml_path, Path(__file__))

    def test_sum_different_simple_format_lengths_exception(self):
        sf1 = SimpleFormat(quarter_durations=[3, 2, 3], midis=[65, 64, 63])
        sf2 = SimpleFormat(quarter_durations=[3, 2, 1], midis=[63, 64, 65])
        with self.assertRaises(SimpleFormatException):
            SimpleFormat.sum(sf1, sf2)



from unittest import TestCase
import uuid
from pathlib import Path

from musicscore import QuarterDurationIsNotWritable, Part, QuarterDuration, Chord, Score
from musicscore.exceptions import ChordTestError
from musicscore.tests.util import create_test_xml_paths
from musicscore.tests.util_subdivisions import generate_subdivsion_test_patterns

path = Path(__file__)


class TestWritingSubdivisions(TestCase):
    def test_generate_patterns_1(self):
        expected = [
            (1, 4),
            (2, 3),
            (3, 2),
            (4, 1),
            (1, 1, 3),
            (1, 2, 2),
            (1, 3, 1),
            (2, 1, 2),
            (2, 2, 1),
            (3, 1, 1),
        ]
        assert list(generate_subdivsion_test_patterns(5)) == expected

    def test_generate_patterns_2(self):
        expected = [
            (1, 5),
            (2, 4),
            (3, 3),
            (4, 2),
            (5, 1),
            (1, 1, 4),
            (1, 2, 3),
            (1, 3, 2),
            (1, 4, 1),
            (2, 1, 3),
            (2, 2, 2),
            (2, 3, 1),
            (3, 1, 2),
            (3, 2, 1),
            (4, 1, 1),
        ]
        assert (
            list(generate_subdivsion_test_patterns(6, remove_larger_subdivision=False))
            == expected
        )
        expected = [
            (1, 5),
            (5, 1),
            (1, 1, 4),
            (1, 2, 3),
            (1, 3, 2),
            (1, 4, 1),
            (2, 1, 3),
            (2, 3, 1),
            (3, 1, 2),
            (3, 2, 1),
            (4, 1, 1),
        ]
        assert list(generate_subdivsion_test_patterns(6, True)) == expected

    def write_and_test_subdivision(self, subdivision):
        for pattern in generate_subdivsion_test_patterns(
            subdivision, remove_larger_subdivision=True
        ):
            p = Part(f"p{uuid.uuid4()}")
            p.add_measure([1, 4])
            qds = [QuarterDuration(x, subdivision) for x in pattern]
            [p.add_chord(Chord(71, qd)) for qd in qds]
            measure = p.get_children()[-1]
            try:
                measure.finalize()
                for ch in measure.get_chords():
                    try:
                        ch.check_printed_duration()
                        ch.check_number_of_beams()
                    except ChordTestError as err:
                        print(qds)
                        print([ch.quarter_duration for ch in measure.get_chords()])
                        print(err)

            except QuarterDurationIsNotWritable as err:
                print(qds)
                print(err)

    def test_write_quarter_upto_7(self):
        for subdivision in range(3, 8):
            self.write_and_test_subdivision(subdivision)

    def test_write_quarter_32nd(self):
        self.write_and_test_subdivision(8)

    def test_write_quarter_9(self):
        self.write_and_test_subdivision(9)

    def test_write_quarter_10(self):
        self.write_and_test_subdivision(10)

    def test_write_quarter_11(self):
        self.write_and_test_subdivision(11)

    def test_write_quarter_12(self):
        self.write_and_test_subdivision(12)

    def test_write_quarter_13(self):
        self.write_and_test_subdivision(13)

    def test_write_quarter_14(self):
        self.write_and_test_subdivision(14)

    def test_write_quarter_15(self):
        self.write_and_test_subdivision(15)

    def test_write_quarter_16(self):
        self.write_and_test_subdivision(16)

    def test_write_quarter_32(self):
        self.write_and_test_subdivision(32)


class TestWriteScoreWithSubdivisions(TestCase):
    def write_subdivision_to_file(self, subdivision):
        score = Score()
        p = score.add_part("p1")
        p.add_measure([1, 4])
        for pattern in generate_subdivsion_test_patterns(
            subdivision, remove_larger_subdivision=True
        ):
            chords = [
                Chord(60, qd)
                for qd in [QuarterDuration(x, subdivision) for x in pattern]
            ]
            [ch.add_lyric(x) for ch, x in zip(chords, pattern)]
            [p.add_chord(ch) for ch in chords]
        xml_path = create_test_xml_paths(path, f"{subdivision}")[0]
        score.export_xml(xml_path)

    def test_write_quarter_32nd(self):
        self.write_subdivision_to_file(8)

    def test_write_quarter_9(self):
        self.write_subdivision_to_file(9)

    def test_write_quarter_10(self):
        self.write_subdivision_to_file(10)

    def test_write_quarter_11(self):
        self.write_subdivision_to_file(11)

    def test_write_quarter_12(self):
        self.write_subdivision_to_file(12)

    def test_write_quarter_13(self):
        self.write_subdivision_to_file(13)

    def test_write_quarter_14(self):
        self.write_subdivision_to_file(14)

    def test_write_quarter_15(self):
        self.write_subdivision_to_file(15)

    def test_write_quarter_16(self):
        self.write_subdivision_to_file(16)

    def test_write_quarter_32(self):
        self.write_subdivision_to_file(32)



from unittest import TestCase, skip

from musicscore.chord import Chord
from musicscore.exceptions import ScoreHasNoPartsError, ScoreMultiMeasureRestError
from musicscore.layout import StaffLayout
from musicscore.measure import Measure
from musicscore.part import Part
from musicscore.score import Score, TITLE, SUBTITLE
from musicxml import XMLNote


class TestScore(TestCase):
    def test_score_version(self):
        s = Score(version=3.1)
        assert s.version == "3.1"
        s = Score()
        assert s.version == "4.0"

    def test_score_init(self):
        s = Score()
        p = s.add_child(Part("p1"))
        p.add_child(Measure(1))
        assert s.xml_part_list.find_child("XMLScorePart") == p.score_part.xml_object
        assert s.find_child("XMLPart") == p.xml_object
        assert s.xml_defaults is not None
        s.xml_object._final_checks()

    def test_score_without_part(self):
        s = Score()
        with self.assertRaises(ScoreHasNoPartsError):
            s.finalize()

    def test_get_chords(self):
        s = Score()
        p = s.add_child(Part("p1"))
        m = p.add_child(Measure(1))
        m._add_chord(Chord(midis=60, quarter_duration=1))
        m._add_chord(Chord(midis=60, quarter_duration=1))
        m._add_chord(Chord(midis=60, quarter_duration=1))
        m._add_chord(Chord(midis=60, quarter_duration=1))
        assert len(s.get_chords()) == 4
        assert type(s.get_chords()[0]) is Chord

    def test_score_title(self):
        s = Score(title="A Nice Piece of Music")
        credits_ = s.xml_object.find_children("XMLCredit")
        assert len(credits_) == 1
        assert credits_[0].page == 1
        assert credits_[0].xml_credit_type.value_ == "title"
        title_words = credits_[0].xml_credit_words
        assert title_words.value_ == "A Nice Piece of Music"
        assert title_words.font_size == TITLE["font_size"]
        assert title_words.justify == "center"
        assert title_words.valign == "top"
        assert title_words.default_x == TITLE["default_x"]["A4"]["portrait"]
        assert title_words.default_y == TITLE["default_y"]["A4"]["portrait"]

        s.title = "Some other Title"
        assert title_words.value_ == "Some other Title"

        s.title = None
        assert len(s.xml_object.find_children("XMLCredit")) == 0

        s.title = "New Title"
        assert len(s.xml_object.find_children("XMLCredit")) == 1

    def test_score_subtitle(self):
        s = Score(title="A Nice Piece of Music", subtitle="No. 2")
        credits_ = s.xml_object.find_children("XMLCredit")
        assert len(credits_) == 2
        assert credits_[1].page == 1
        assert credits_[1].xml_credit_type.value_ == "subtitle"
        subtitle_words = credits_[1].xml_credit_words
        assert subtitle_words.value_ == "No. 2"
        assert subtitle_words.font_size == SUBTITLE["font_size"]
        assert subtitle_words.halign == "center"
        assert subtitle_words.valign == "top"
        assert subtitle_words.default_x == SUBTITLE["default_x"]["A4"]["portrait"]
        assert subtitle_words.default_y == SUBTITLE["default_y"]["A4"]["portrait"]

        s.subtitle = "Some other Subtitle"
        assert subtitle_words.value_ == "Some other Subtitle"

        s.subtitle = None
        assert len(s.xml_object.find_children("XMLCredit")) == 1

        s.subtitle = "New Subtitle"
        assert len(s.xml_object.find_children("XMLCredit")) == 2

    @skip
    def test_score_composer(self):
        self.fail("Incomplete")

    @skip
    def test_score_add_other_credits(self):
        self.fail("Incomplete")

    def test_score_scaling(self):
        s = Score()
        assert s.scaling.score == s
        assert s.scaling.tenths == 40
        assert s.scaling.millimeters == 7.2319
        assert (
            s.xml_object.xml_defaults.xml_scaling.xml_tenths.value_ == s.scaling.tenths
        )
        assert (
            s.xml_object.xml_defaults.xml_scaling.xml_millimeters.value_
            == s.scaling.millimeters
        )
        s.scaling.tenths = 50
        assert (
            s.xml_object.xml_defaults.xml_scaling.xml_tenths.value_ == s.scaling.tenths
        )
        assert (
            s.xml_object.xml_defaults.xml_scaling.xml_millimeters.value_
            == s.scaling.millimeters
        )

    def test_score_page_layout(self):
        s = Score()
        assert s.page_layout.parent == s
        assert s.page_layout.scaling == s.scaling
        assert s.xml_object.xml_defaults.xml_page_layout.xml_page_height.value_ == 1643
        assert (
            s.xml_object.xml_defaults.xml_page_layout.xml_page_margins.xml_left_margin.value_
            == 140
        )

        s.page_layout.orientation = "landscape"
        assert s.xml_object.xml_defaults.xml_page_layout.xml_page_height.value_ == 1161
        assert (
            s.xml_object.xml_defaults.xml_page_layout.xml_page_margins.xml_left_margin.value_
            == 111
        )

    def test_score_system_layout(self):
        s = Score()
        assert s.system_layout.parent == s
        assert (
            s.xml_object.xml_defaults.xml_system_layout.xml_system_margins.xml_left_margin.value_
            == 0
        )
        assert (
            s.xml_object.xml_defaults.xml_system_layout.xml_system_margins.xml_right_margin.value_
            == 0
        )
        assert (
            s.xml_object.xml_defaults.xml_system_layout.xml_system_distance.value_
            == 117
        )
        assert (
            s.xml_object.xml_defaults.xml_system_layout.xml_top_system_distance.value_
            == 117
        )
        s.system_layout.margins.left = 10
        assert (
            s.xml_object.xml_defaults.xml_system_layout.xml_system_margins.xml_left_margin.value_
            == 10
        )
        s.system_layout.top_system_distance = 100
        assert (
            s.xml_object.xml_defaults.xml_system_layout.xml_top_system_distance.value_
            == 100
        )

    def test_score_staff_layout(self):
        s = Score()
        assert s.staff_layout is None
        s.staff_layout = StaffLayout()
        assert (
            s.xml_object.xml_defaults.xml_staff_layout.xml_staff_distance.value_ == 80
        )
        s.staff_layout.staff_distance = 100
        assert (
            s.xml_object.xml_defaults.xml_staff_layout.xml_staff_distance.value_ == 100
        )

    def test_score_identification_and_new_system(self):
        s = Score()
        assert s.xml_object.xml_identification is not None
        assert s.xml_object.xml_identification.xml_encoding is not None
        all_supports = s.xml_object.xml_identification.xml_encoding.find_children(
            "XMLSupports"
        )
        assert len(all_supports) == 3
        expected = """<identification>
    <encoding>
      <supports element="accidental" type="yes" />
      <supports element="beam" type="yes" />
      <supports element="stem" type="yes" />
    </encoding>
  </identification>
"""
        assert s.xml_object.xml_identification.to_string() == expected
        s.new_system = True
        expected = """<identification>
    <encoding>
      <supports element="accidental" type="yes" />
      <supports element="beam" type="yes" />
      <supports element="stem" type="yes" />
      <supports attribute="new-system" element="print" type="yes" value="yes" />
    </encoding>
  </identification>
"""
        assert s.xml_object.xml_identification.to_string() == expected
        s.new_system = False
        expected = """<identification>
    <encoding>
      <supports element="accidental" type="yes" />
      <supports element="beam" type="yes" />
      <supports element="stem" type="yes" />
    </encoding>
  </identification>
"""
        assert s.xml_object.xml_identification.to_string() == expected

    def test_group_parts_wrong_number(self):
        score = Score()
        [score.add_part(f"p-{i}") for i in range(1, 3)]
        with self.assertRaises(ValueError):
            score.group_parts(1, 0, 1)
        with self.assertRaises(ValueError):
            score.group_parts(2, 1, 6)
        with self.assertRaises(ValueError):
            score.group_parts(3, 2, 1)

    def test_create_missing_measures(self):
        score = Score()
        parts = [score.add_part(f"p-{i}") for i in range(1, 4)]
        [parts[0].add_chord(Chord(60, 4)) for _ in range(5)]
        [parts[2].add_chord(Chord(60, 4)) for _ in range(3)]
        score.finalize()
        for p in parts:
            assert len(p.get_children()) == 5
        for m in parts[1].get_children():
            assert m.get_chords()[0].is_rest
        for m in parts[2].get_children()[3:]:
            assert m.get_chords()[0].is_rest

    def test_group_parts(self):
        score = Score()
        parts = [score.add_part(f"p-{i}") for i in range(1, 6)]
        for p in parts:
            p.add_chord(Chord(0, 4))

        score.group_parts(1, 2, 4, name="Group 1", symbol="square")
        score.group_parts(2, 3, 4, name="Group 2", symbol="bracket")

        expected = """<part-list>
    <score-part id="p-1">
      <part-name />
    </score-part>
    <part-group number="1" type="start">
      <group-name>Group 1</group-name>
      <group-symbol>square</group-symbol>
      <group-barline>yes</group-barline>
    </part-group>
    <score-part id="p-2">
      <part-name />
    </score-part>
    <part-group number="2" type="start">
      <group-name>Group 2</group-name>
      <group-symbol>bracket</group-symbol>
      <group-barline>yes</group-barline>
    </part-group>
    <score-part id="p-3">
      <part-name />
    </score-part>
    <score-part id="p-4">
      <part-name />
    </score-part>
    <part-group number="2" type="stop" />
    <part-group number="1" type="stop" />
    <score-part id="p-5">
      <part-name />
    </score-part>
  </part-list>
"""
        score.finalize()
        assert score.xml_part_list.to_string() == expected

    def test_set_multiple_measure_rests_one_part(self):
        score = Score()
        with self.assertRaises(ScoreMultiMeasureRestError):
            score.set_multi_measure_rest(1, 4)
        part = score.add_part("p1")
        assert len(part.get_children()) == 0
        # add 4 measures and set multiple rests in measure attributes
        score.set_multi_measure_rest(1, 4)
        assert len(part.get_children()) == 4
        assert (
            part.get_measure(
                1
            ).xml_attributes.xml_measure_style.xml_multiple_rest.value_
            == 4
        )

        # Trying to set multiple rest measures again throws error
        with self.assertRaises(ScoreMultiMeasureRestError):
            score.set_multi_measure_rest(2, 3)

        # add 4 measures manually
        [part.add_chord(Chord(0, 4)) for _ in range(4)]
        assert len(part.get_children()) == 8
        # set multiple measure rests for the last three measures
        score.set_multi_measure_rest(6, 8)
        assert part.get_measure(5).xml_attributes.xml_measure_style is None
        assert (
            part.get_measure(
                6
            ).xml_attributes.xml_measure_style.xml_multiple_rest.value_
            == 3
        )
        # Trying to set multiple rest measures on measures with pitches throws error
        [part.add_chord(Chord(60, 4)) for _ in range(4)]
        with self.assertRaises(ScoreMultiMeasureRestError):
            score.set_multi_measure_rest(9, 10)

        # Check if rest measure attribute is set to yes
        score.finalize()
        for measure in part.get_children():
            if int(measure.number) in [1, 2, 3, 4, 6, 7, 8]:
                assert (
                    measure.xml_object.get_children_of_type(XMLNote)[0].xml_rest.measure
                    == "yes"
                )
            elif int(measure.number) == 5:
                assert (
                    measure.xml_object.get_children_of_type(XMLNote)[0].xml_rest.measure
                    is None
                )
            else:
                assert (
                    measure.xml_object.get_children_of_type(XMLNote)[0].xml_rest is None
                )

    def test_set_multiple_measure_rests_multiple_parts(self):
        score = Score()
        parts = p1, p2 = score.add_part("p1"), score.add_part("p2")
        # add 4 measures and set multiple rests in measure attributes
        score.set_multi_measure_rest(1, 4)
        for p in parts:
            assert len(p.get_children()) == 4
            assert (
                p.get_measure(
                    1
                ).xml_attributes.xml_measure_style.xml_multiple_rest.value_
                == 4
            )

        # add 4 measures manually
        [p1.add_chord(Chord(0, 4)) for _ in range(4)]
        assert len(p1.get_children()) == 8
        assert len(p2.get_children()) == 4
        # set multiple measure rests for the last three measures
        score.set_multi_measure_rest(6, 8)
        assert len(p2.get_children()) == 8
        for p in parts:
            assert p.get_measure(5).xml_attributes.xml_measure_style is None
            assert (
                p.get_measure(
                    6
                ).xml_attributes.xml_measure_style.xml_multiple_rest.value_
                == 3
            )
        # Trying to set multiple rest measures on measures with pitches throws error
        [p1.add_chord(Chord(60, 4)) for _ in range(4)]
        with self.assertRaises(ScoreMultiMeasureRestError):
            score.set_multi_measure_rest(9, 10)
        # Check if rest measure attribute is set to yes
        score.finalize()
        for measure in p1.get_children():
            if int(measure.number) in [1, 2, 3, 4, 6, 7, 8]:
                assert (
                    measure.xml_object.get_children_of_type(XMLNote)[0].xml_rest.measure
                    == "yes"
                )
            elif int(measure.number) in [5]:
                assert (
                    measure.xml_object.get_children_of_type(XMLNote)[0].xml_rest.measure
                    is None
                )
            else:
                assert (
                    measure.xml_object.get_children_of_type(XMLNote)[0].xml_rest is None
                )

        for measure in p2.get_children():
            if int(measure.number) in [1, 2, 3, 4, 6, 7, 8]:
                assert (
                    measure.xml_object.get_children_of_type(XMLNote)[0].xml_rest.measure
                    == "yes"
                )
            else:
                assert (
                    measure.xml_object.get_children_of_type(XMLNote)[0].xml_rest.measure
                    is None
                )

    def test_barline(self):
        score = Score()
        parts = [score.add_part(f"p-{i}") for i in range(1, 4)]
        [parts[0].add_chord(Chord(60, 4)) for _ in range(5)]
        [parts[2].add_chord(Chord(60, 4)) for _ in range(3)]
        parts[2].get_measure(2).set_barline(style="light-light")
        parts[0].get_measure(2).set_barline(style="light-heavy")
        parts[0].get_measure(4).set_barline(style="heavy-heavy")
        score.finalize()
        for p in parts:
            assert p.get_measure(2).get_barline().xml_bar_style.value_ == "light-heavy"
            assert p.get_measure(4).get_barline().xml_bar_style.value_ == "heavy-heavy"

    def test_last_measure_barline(self):
        score = Score()
        parts = [score.add_part(f"p-{i}") for i in range(1, 6)]
        for p in parts:
            for _ in range(3):
                p.add_chord(Chord(0, 4))
        score.finalize()
        for p in parts:
            assert p.get_children()[-1].xml_barline.location == "right"
            assert (
                p.get_children()[-1].xml_barline.xml_bar_style.value_ == "light-heavy"
            )

    def test_last_measure_barline_already_set(self):
        score = Score()
        parts = [score.add_part(f"p-{i}") for i in range(1, 6)]
        for p in parts:
            for _ in range(3):
                p.add_chord(Chord(0, 4))
        parts[0].get_children()[-1].set_barline(style="light-light")
        score.finalize()
        for p in parts:
            assert p.get_children()[-1].xml_barline.location == "right"
            assert (
                p.get_children()[-1].xml_barline.xml_bar_style.value_ == "light-light"
            )



class XMLWrapper:
    """
    XMLWrapper contains an xml object at its core. It is the place where all the intuitive stuff is translated to attributes and children of
    xml objects.

    All attributes and properties which are not listed in _ATTRIBUTES will be set to or get from core xml object.
    """

    _ATTRIBUTES = {}

    XMLClass = None

    @property
    def xml_object(self) -> "XMLElement":
        """
        :obj:`~musicscore.xmlwrapper.XMLWrapper` method

        This musicscore class has an equivalent ``MusicXML`` class which can be reached via this property.

        If an attribute is not found directly in the main class, it wll be passed on to ``__get_attribute__`` or ``__set__attribute__`` methods of ``self.xml_object``.
        It can also use the short cut attributes of :obj:`~musicxml.xmlelement.xmlelement.XMLElement` to get or set the first child.

        :return: wrapped MusicXML element of type :obj:`~XMLClass`
        """
        return self._xml_object

    def to_string(self, *args, **kwargs) -> str:
        """
        :obj:`~musicscore.xmlwrapper.XMLWrapper` method

        Calls :obj:`~musicxml.xmlelement.xmlelement.XMLElement.to_string` method of self.xml_object

        :return: musicxml snippet
        """
        if self.xml_object:
            return self.xml_object.to_string(*args, **kwargs)
        else:
            raise ValueError(f"{self.__class__.__name__} has no xml object.")

    def __setattr__(self, key, value):
        attributes = self._ATTRIBUTES
        try:
            if self._TREE_ATTRIBUTES:
                attributes.union(self._TREE_ATTRIBUTES)
        except AttributeError:
            pass
        if key[0] != "_" and (
            "_xml_object" in self.__dict__
            and key not in attributes
            and key not in self.__dict__
        ):
            setattr(self._xml_object, key, value)
        else:
            super().__setattr__(key, value)

    def __getattr__(self, item):
        if item == "_TREE_ATTRIBUTES":
            raise AttributeError
        if item == "xml_object":
            return super().__getattribute__(item)
        try:
            return self._xml_object.__getattribute__(item)
        except AttributeError:
            try:
                return self._xml_object.__getattr__(item)
            except AttributeError:
                return super().__getattribute__(item)



from fractions import Fraction
from pathlib import Path
from musicscore.chord import Chord
from musicscore.score import Score
from musicscore.tests.util import XMLTestCase

path = Path(__file__)


class DebuggingTests(XMLTestCase):
    def setUp(self):
        self.score = Score()
        return super().setUp()

    def test_septuplets_with_gracenote(self):
        durations = [
            Fraction(0, 1),
            Fraction(3, 7),
            Fraction(1, 7),
            Fraction(2, 7),
            Fraction(1, 7),
        ]
        part = self.score.add_part("part-1")
        part.add_measure(time=(1, 4))
        for duration in durations:
            part.add_chord(Chord(60, duration))

        with self.file_path(path, "septuplets_with_gracenote") as xml_path:
            self.score.export_xml(xml_path)

    def test_grouping_problems(self):
        durations = [
            Fraction(19, 675),
            Fraction(49, 646),
            Fraction(30, 791),
            Fraction(15, 791),
            Fraction(45, 791),
            Fraction(16, 375),
            Fraction(13, 914),
            Fraction(45, 791),
            Fraction(16, 225),
            Fraction(19, 668),
            Fraction(32, 675),
            Fraction(16, 135),
            Fraction(16, 675),
            Fraction(16, 225),
            Fraction(64, 675),
            Fraction(9, 949),
            Fraction(15, 791),
            Fraction(19, 668),
            Fraction(30, 791),
            Fraction(32, 675),
            Fraction(15, 791),
            Fraction(13, 914),
            Fraction(16, 675),
            Fraction(9, 949),
            Fraction(1, 211),
        ]
        # print(sum(durations).limit_denominator(1000))
        self.score.get_quantized = True
        part = self.score.add_part("part-1")
        for duration in durations:
            part.add_chord(Chord(60, duration))
        # print([ch.quarter_duration for ch in part.get_chords()])
        with self.file_path(path, "grouping_problems") as xml_path:
            self.score.export_xml(xml_path)

    def test_update_accidentals_with_rest(self):
        score = Score()
        part = score.add_part("p1")
        qds = [
            1 / 5,
            4 / 15,
            1 / 3,
            8 / 21,
            32 / 105,
            16 / 105,
            8 / 35,
            4 / 3,
            16 / 15,
            64 / 147,
            256 / 735,
            128 / 735,
            64 / 245,
            8 / 35,
            32 / 105,
            8 / 21,
            32 / 21,
            64 / 105,
            4 / 9,
            4 / 15,
            16 / 45,
            16 / 45,
            64 / 315,
            32 / 105,
            128 / 315,
            32 / 63,
            16 / 9,
            32 / 45,
            32 / 27,
            128 / 945,
            64 / 189,
            64 / 315,
            256 / 945,
        ]
        for qd in qds:
            chord = Chord(61, quarter_duration=qd)
            if qd in [16 / 105, 128 / 735, 64 / 105, 64 / 315, 32 / 45, 128 / 945]:
                chord.midis = 0
            part.add_chord(chord)
        part.get_quantized = True
        part.finalize()

    def test_tied_unwritables(self):
        score = Score()
        part = score.add_part("p1")
        part.add_measure(time=(2, 4))
        qd_patterns = [[2 / 7, 12 / 7], [1 / 8, 15 / 8], [1 / 6, 11 / 6]]
        for pattern in qd_patterns:
            for qd in pattern:
                part.add_chord(Chord(61, quarter_duration=qd))

        with self.file_path(path, "tied_unwritables") as xml_path:
            score.export_xml(xml_path)



from unittest import TestCase

from musicscore.exceptions import MusicTreeTypeError
from musicscore.midi import Midi
from musicscore.accidental import Accidental


class TestAccidental(TestCase):
    def test_get_pitch_parameters(self):
        a = Accidental()
        assert a.get_pitch_parameters(midi_value=None) is None
        assert a.get_pitch_parameters(midi_value=0) is None
        assert a.get_pitch_parameters(midi_value=61) == ("C", 1, 4)
        a.mode = "sharp"
        assert a.get_pitch_parameters(midi_value=61) == ("C", 1, 4)
        a.mode = "flat"
        assert a.get_pitch_parameters(midi_value=61) == ("D", -1, 4)
        a.mode = "enharmonic"
        assert a.get_pitch_parameters(midi_value=61) == ("D", -1, 4)
        a.mode = "force-flat"
        assert a.get_pitch_parameters(midi_value=61) == ("D", -1, 4)
        a.mode = "force-sharp"
        assert a.get_pitch_parameters(midi_value=61) == ("B", 2, 3)

        midi = Midi(61, accidental=a)
        assert a.get_pitch_parameters() == ("B", 2, 3)
        assert midi.accidental.get_pitch_parameters() == ("B", 2, 3)

    def test_accidental_sign(self):
        a = Accidental()
        assert a.sign is None
        midi = Midi(61)
        assert midi.accidental.sign == "sharp"
        midi.accidental.mode = "flat"
        assert midi.accidental.sign == "flat"
        midi.accidental.mode = "force-sharp"
        assert midi.accidental.sign == "double-sharp"
        midi.value = 60
        assert midi.accidental.sign == "sharp"
        midi.accidental.mode = "flat"
        assert midi.accidental.sign == "natural"
        midi.accidental.mode = "enharmonic"
        assert midi.accidental.sign == "natural"
        midi.accidental.mode = "force-flat"
        assert midi.accidental.sign == "flat-flat"

        midi = Midi(61)
        assert midi.accidental.sign == "sharp"
        midi.value = 60
        assert midi.accidental.sign == "natural"

    def test_get_xml_accidental(self):
        _ = Accidental()
        midi = Midi(61)
        midi.accidental.show = True
        expected = """<accidental>sharp</accidental>
"""
        assert midi.accidental.xml_object.to_string() == expected
        midi.accidental.mode = "flat"
        expected = """<accidental>flat</accidental>
"""
        assert midi.accidental.xml_object.to_string() == expected
        midi.accidental.mode = "force-sharp"
        expected = """<accidental>double-sharp</accidental>
"""
        assert midi.accidental.xml_object.to_string() == expected
        midi.value = 60
        expected = """<accidental>sharp</accidental>
"""
        assert midi.accidental.xml_object.to_string() == expected
        midi.value = 0
        assert midi.accidental.xml_object is None

    def test_accidental_show_mode(self):
        midi = Midi(60)
        assert midi.accidental.show is None
        assert midi.accidental.sign == "natural"
        midi.accidental.show = True
        assert midi.accidental.xml_object.value_ == "natural"
        midi.accidental.show = False
        assert midi.accidental.xml_object is None

        midi.accidental.show = True
        midi.value = 61
        assert midi.accidental.sign == "sharp"
        assert midi.accidental.xml_object.value_ == "sharp"
        midi.accidental.mode = "flat"
        assert midi.accidental.xml_object.value_ == "flat"
        midi.accidental.show = False
        assert midi.accidental.xml_object is None

    def test_accidental_midi(self):
        midi = Midi(61)
        assert midi.accidental.sign == "sharp"
        midi.accidental.show = True
        assert midi.accidental.xml_object.value_ == "sharp"

    def test_accidental_up_midi(self):
        m = Midi(70)
        assert m.accidental.up == m

    def test_accidental_copy(self):
        a = Accidental(mode="sharp", show=True)
        copied = a.__copy__()
        assert a != copied
        assert a.xml_object != copied.xml_object
        assert a.xml_object.value_ == copied.xml_object.value_
        assert a.mode == copied.mode
        assert a.show == copied.show

    def test_add_child(self):
        a = Accidental()
        with self.assertRaises(NotImplementedError):
            a.add_child(Accidental())

    def test_get_children(self):
        a = Accidental()
        assert a.get_children() == []

    def test_get_leaves(self):
        a = Accidental()
        assert a.get_leaves() == [a]

    def test_not_implemented_inherited_methods(self):
        a = Accidental()
        with self.assertRaises(MusicTreeTypeError):
            a.get_chord()
        with self.assertRaises(MusicTreeTypeError):
            a.get_beat()
        with self.assertRaises(MusicTreeTypeError):
            a.get_voice()
        with self.assertRaises(MusicTreeTypeError):
            a.get_staff()
        with self.assertRaises(MusicTreeTypeError):
            a.get_measure()
        with self.assertRaises(MusicTreeTypeError):
            a.get_part()
        with self.assertRaises(MusicTreeTypeError):
            a.get_beats()
        with self.assertRaises(MusicTreeTypeError):
            a.get_chords()



from unittest import TestCase
from musicscore import Part
from musicscore.clef import (
    Clef,
    TrebleClef,
    BassClef,
    TenorClef,
    AltoClef,
    PercussionClef,
)


class TestClef(TestCase):
    def test_default_clef(self):
        c = Clef()
        expected = """<clef>
  <sign>G</sign>
  <line>2</line>
</clef>
"""
        assert c.to_string() == expected

    def test_clef_copy(self):
        c = Clef("F", 2, show=False, octave_change=-1)
        copied = c.__copy__()
        assert copied != c
        assert copied.xml_object != c.xml_object
        assert copied.sign == c.sign
        assert copied.line == c.line
        assert copied.show == c.show
        assert copied.octave_change == c.octave_change

    def test_global_clefs(self):
        c = TrebleClef()
        assert c.sign == "G"
        assert c.line == 2
        c = BassClef()
        assert c.sign == "F"
        assert c.line == 4
        c = TenorClef()
        assert c.sign == "C"
        assert c.line == 4
        c = AltoClef()
        assert c.sign == "C"
        assert c.line == 3
        c = PercussionClef()
        assert c.sign == "percussion"
        assert c.line is None

        c = TrebleClef(octave_change=1)
        expected = """<clef>
  <sign>G</sign>
  <line>2</line>
  <clef-octave-change>1</clef-octave-change>
</clef>
"""
        assert c.to_string() == expected

    def test_staff_clef(self):
        clefs = [
            None,
            TrebleClef(),
            AltoClef(),
            TenorClef(),
            BassClef(),
            PercussionClef(),
            TrebleClef(octave_change=-1),
            Clef(sign="F", line=3),
        ]

        part = Part("p1")
        for c in clefs:
            m = part.add_measure()
            m.get_staff(1).clef = c
        for m, c in zip(part.get_children(), clefs):
            if c:
                assert m.clefs == [c]
        part.finalize()



from unittest import TestCase
from musicxml.xmlelement.xmlelement import XMLScaling

from musicscore.layout import PageLayout, SystemLayout, StaffLayout, Scaling
from musicscore.score import Score


class TestScaling(TestCase):
    def test_scaling_init(self):
        sc = Scaling()
        assert sc.tenths == 40
        assert sc.millimeters == 7.2319
        assert isinstance(sc.xml_object, XMLScaling)
        assert sc.xml_object.xml_tenths.value_ == sc.tenths
        assert sc.xml_object.xml_millimeters.value_ == sc.millimeters

        sc.tenths = 50
        assert sc.xml_object.xml_tenths.value_ == sc.tenths


class TestPageLayout(TestCase):
    def setUp(self):
        self.score = Score()

    def test_page_layout_default(self):
        pl = PageLayout()
        self.score.page_layout = pl
        assert pl.size == "A4"
        assert pl.orientation == "portrait"
        assert pl.xml_page_height.value_ == 1643
        assert pl.xml_page_width.value_ == 1161
        assert pl.xml_page_margins.type == "both"
        assert pl.xml_page_margins.xml_left_margin.value_ == 140
        assert pl.xml_page_margins.xml_right_margin.value_ == 70
        assert pl.xml_page_margins.xml_top_margin.value_ == 70
        assert pl.xml_page_margins.xml_bottom_margin.value_ == 70

        pl.orientation = "landscape"
        assert pl.xml_page_height.value_ == 1161
        assert pl.xml_page_width.value_ == 1643
        assert pl.xml_page_margins.type == "both"
        assert pl.xml_page_margins.xml_left_margin.value_ == 111
        assert pl.xml_page_margins.xml_right_margin.value_ == 70
        assert pl.xml_page_margins.xml_top_margin.value_ == 70
        assert pl.xml_page_margins.xml_bottom_margin.value_ == 70

        pl.size = "A3"
        pl.orientation = "portrait"
        assert pl.xml_page_height.value_ == 2323
        assert pl.xml_page_width.value_ == 1643
        assert pl.xml_page_margins.type == "both"
        assert pl.xml_page_margins.xml_left_margin.value_ == 111
        assert pl.xml_page_margins.xml_right_margin.value_ == 70
        assert pl.xml_page_margins.xml_top_margin.value_ == 70
        assert pl.xml_page_margins.xml_bottom_margin.value_ == 70

        pl.orientation = "landscape"
        assert pl.xml_page_height.value_ == 1643
        assert pl.xml_page_width.value_ == 2323
        assert pl.xml_page_margins.type == "both"
        assert pl.xml_page_margins.xml_left_margin.value_ == 111
        assert pl.xml_page_margins.xml_right_margin.value_ == 70
        assert pl.xml_page_margins.xml_top_margin.value_ == 70
        assert pl.xml_page_margins.xml_bottom_margin.value_ == 70

        pl.scaling.tenths = 50
        assert self.score.page_layout == pl
        assert self.score.scaling.score == self.score
        assert self.score.scaling.tenths == 50
        assert pl.xml_page_height.value_ == round(1643 * 5 / 4)


class TestSystemLayout(TestCase):
    def test_style_layout_default(self):
        sl = SystemLayout()
        assert sl.xml_system_distance.value_ == 117
        assert sl.xml_top_system_distance.value_ == 117
        assert sl.xml_system_margins.xml_left_margin.value_ == 0
        assert sl.xml_system_margins.xml_right_margin.value_ == 0


class TestStaffLayout(TestCase):
    def test_staff_layout_default(self):
        sl = StaffLayout()
        assert sl.xml_staff_distance.value_ == 80



"""
All types of octave shifts (15ma, 15mb, 8va, 8vb)
"""

from pathlib import Path

from musicscore import Score, Chord, C, A, B
from unittest import TestCase
from musicscore.util import octave_chords


class TestLily33d(TestCase):
    def test_lily_33d_Spanners_OctaveShifts(self):
        score = Score()
        part = score.add_part("p1")
        midis = [A(4), C(5), A(6), C(3), B(2), A(5), A(5), B(3), C(4)]
        quarter_durations = 7 * [0.5] + 2 * [0.25]
        chords = [Chord(m, d) for m, d in zip(midis, quarter_durations)]

        octave_chords(chords[2], size=15)
        octave_chords(chords[3:5], type="up", size=15)
        octave_chords(chords[5:7], type="down", size=8)
        octave_chords(chords[7:], type="up", size=8)

        [part.add_chord(ch) for ch in chords]

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



import copy
from math import log2
from typing import Optional, Union

from musicscore import STANDARD, ENHARMONIC, FLAT, SHARP, FORCEFLAT, FORCESHARP
from musicscore.exceptions import AlreadyFinalizedError
from musicxml.xmlelement.xmlelement import *  # type: ignore

from musicscore.accidental import Accidental
from musicscore.musictree import MusicTree

__all__ = [
    "Midi",
    "MidiNote",
    "C",
    "D",
    "E",
    "F",
    "G",
    "A",
    "B",
    "midi_to_frequency",
    "frequency_to_midi",
    "get_accidental_mode",
]


def midi_to_frequency(midi, a4=440):
    try:
        midi = midi.value
    except AttributeError:
        pass

    f = 2 ** ((midi - 69) / 12) * a4
    return f


def frequency_to_midi(frequency, a4=440):
    m = 69 + 12 * log2(frequency / a4)
    return m


def get_accidental_mode(
    midi_value: Union[float, int], accidental_sign: Optional[str] = None
) -> str:
    """
    :param midi_value: a valid midi value in half steps (int or float)
    :param accidental_sign: ``double-flat``, ``flat-flat``, ``bb``, ``ff`` – ``three-quarters-flat`` – ``flat``, ``b``, ``f`` – ``quarter-flat`` – ``None``, ``natural`` – ``quarter-sharp`` – ``sharp``, ``#``, ``s`` – ``three-quarters-sharp`` – ``double-sharp``, ``sharp-sharp``, ``x``, ``##``, ``ss``
    :return: accidental_mode: ``standard``, ``enharmonic``, ``flat``, ``sharp``, ``force-flat``, ``force-sharp``
    """
    # midi value is valid
    Midi(midi_value)
    if accidental_sign in [None, "natural"]:
        accidental_value = 0
    elif accidental_sign in ["quarter-sharp"]:
        accidental_value = 0.5
    elif accidental_sign in ["sharp", "#", "s"]:
        accidental_value = 1
    elif accidental_sign in ["three-quarters-sharp"]:
        accidental_value = 1.5
    elif accidental_sign in ["double-sharp", "sharp-sharp", "x", "##", "ss"]:
        accidental_value = 2
    elif accidental_sign in ["quarter-flat"]:
        accidental_value = -0.5
    elif accidental_sign in ["flat", "b", "f"]:
        accidental_value = -1
    elif accidental_sign in ["three-quarters-flat"]:
        accidental_value = -1.5
    elif accidental_sign in ["double-flat", "flat-flat", "bb", "ff"]:
        accidental_value = -2
    else:
        raise NotImplementedError(accidental_sign)

    if STANDARD[midi_value % 12][1] == accidental_value:
        return "standard"
    elif ENHARMONIC[midi_value % 12][1] == accidental_value:
        return "enharmonic"
    elif FLAT[midi_value % 12][1] == accidental_value:
        return "flat"
    elif SHARP[midi_value % 12][1] == accidental_value:
        return "sharp"
    elif FORCEFLAT[midi_value % 12][1] == accidental_value:
        return "force-flat"
    elif FORCESHARP[midi_value % 12][1] == accidental_value:
        return "force-sharp"


class Midi(MusicTree):
    """
    Parent type: :obj:`~musicscore.note.Note`

    Child type: :obj:`~musicscore.accidental.Accidental`

    Midi is the representation of a Pitch with its midi value, and accidental sign. This object is used to create a Chord
    consisting of one or more pitches. The midi representation of a rest is a Midi object with value 0.
    """

    def __init__(
        self,
        value: Union[float, int],
        accidental: Optional[Accidental] = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._value = None
        self._accidental = None
        self._notehead = None
        self._pitch_or_rest = None
        self._parent_chord = None
        self._parent_note = None
        self._ties = set()

        self._staff_number = None

        self.value = value
        self.accidental = accidental

    def _set_parent_chord(self, value):
        if value is not None and "Chord" not in [
            cls.__name__ for cls in value.__class__.__mro__
        ]:
            raise TypeError
        self._parent_chord = value
        # self._parent = value

    def _update_parent_note(self):
        if self.parent_note:
            self.parent_note._update_xml_pitch_or_rest()
            self.parent_note._update_xml_accidental()

    def _update_pitch_parameters(self):
        pitch = self.get_pitch_or_rest()
        if isinstance(pitch, XMLPitch):
            if not self.accidental.get_pitch_parameters()[1]:
                if pitch.xml_alter:
                    pitch.remove(pitch.xml_alter)
                pitch.xml_step, pitch.xml_octave = (
                    self.accidental.get_pitch_parameters()[0],
                    self.accidental.get_pitch_parameters()[2],
                )
            else:
                pitch.xml_step, pitch.xml_alter, pitch.xml_octave = (
                    self.accidental.get_pitch_parameters()
                )
        else:
            raise TypeError

    def _update_pitch_or_rest(self):
        if self._pitch_or_rest is None:
            if self.value == 0:
                self._pitch_or_rest = XMLRest()
            else:
                self._pitch_or_rest = XMLPitch()
        else:
            if self.value == 0:
                if not isinstance(self._pitch_or_rest, XMLRest):
                    self._pitch_or_rest = XMLRest()
                    self._update_parent_note()
                else:
                    pass
            else:
                if not isinstance(self._pitch_or_rest, XMLPitch):
                    self._pitch_or_rest = XMLPitch()
                    self._update_parent_note()
                if self.accidental:
                    self.accidental._update_xml_object()
                    self._update_pitch_parameters()
        if self.up:
            self.up._update_xml_pitch_or_rest()

    # //public properties
    @property
    def accidental(self) -> "Accidental":
        """
        Set or get :obj:`~musicscore.accidental.Accidental` object associated with this midi. If it is set to ``None`` an :obj:`~musicscore.accidental.Accidental` object will be created.
        """
        return self._accidental

    @accidental.setter
    def accidental(self, value):
        if not value:
            value = Accidental()
        elif not isinstance(value, Accidental):
            raise TypeError(
                f"accidental.value must be of type Accidental not {type(value)}"
            )
        self._accidental = value
        if value:
            self._accidental.parent_midi = self

    @property
    def is_tied_to_next(self) -> bool:
        """
        :return: ``True`` if this midi adds a :obj:`~musicxml.xmlelement.xmlelement.XMLTie` object of type ``start`` :obj:`~musicscore.note.Note`.
        """
        if "start" in self._ties:
            return True
        else:
            return False

    @property
    def is_tied_to_previous(self):
        """
        :return: ``True`` if this midi adds a :obj:`~musicxml.xmlelement.xmlelement.XMLTie` object of type ``stop`` :obj:`~musicscore.note.Note`.

        """
        if "stop" in self._ties:
            return True
        else:
            return False

    @property
    def notehead(self) -> Optional["XMLNotehead"]:
        """
        Set or get notedhead property. This can be ``None`` or an :obj:`~musicxml.xmlelement.xmlelement.XMLNotehead` object. It is possible to set this property with a valid notehead value of type str. For permitted values see :obj:`~musicxml.xmlelement.xmlelement.XMLNotehead`.
        """
        return self._notehead

    @notehead.setter
    def notehead(self, val):
        if self.parent_note:
            raise AlreadyFinalizedError("Cannot change notehead after finalizing.")

        if val is None:
            self._notehead = None
        elif not isinstance(val, XMLNotehead):
            self._notehead = XMLNotehead(val)
        else:
            self._notehead = val

    @property
    def octave(self) -> int:
        """
        :return: octave number of this midi.
        """
        return int(self.value / 12) - 1

    @property
    def parent_chord(self) -> "Chord":
        """
        Set or get parent :obj:`~musicscore.chord.Chord` object.
        """
        return self._parent_chord

    @property
    def parent_note(self) -> "Note":
        """
        Set or get parent :obj:`~musicscore.note.Note` object.
        """
        return self._parent_note

    @parent_note.setter
    def parent_note(self, value):
        if value is not None and "Note" not in [
            cls.__name__ for cls in value.__class__.__mro__
        ]:
            raise TypeError
        self._parent_note = value
        self._parent = value

    @property
    def value(self) -> Union[float, int]:
        """
        Set and get value of midi. A valid value must be of type ``float`` or ``int`` and can be between ``12`` and ``127``.
        """
        return self._value

    @value.setter
    def value(self, v):
        if not isinstance(v, float) and not isinstance(v, int):
            raise TypeError(f"Midi.value must be of type float or int not{type(v)}")
        if v != 0 and (v < 12 or v > 127):
            raise ValueError(
                f"Midi.value {v} can be zero for a rest or must be in a range between 12 and 127 inclusively"
            )
        self._value = v
        self._update_pitch_or_rest()

    @property
    def name(self) -> str:
        """
        :return: a string like ``C#3`` consisting of ``step``, ``accidental sign`` (or value) and ``octave``. Midi with value 0 returns ``rest`` as its name.
        """
        if self.value == 0:
            return "rest"

        pitch_step = self.accidental.get_pitch_parameters()[1]

        if not pitch_step:
            accidental = ""
        elif pitch_step == -1.5:
            accidental = "b-"
        elif pitch_step == -1:
            accidental = "b"
        elif pitch_step == -0.5:
            accidental = "-"
        elif pitch_step == 0.5:
            accidental = "+"
        elif pitch_step == 1:
            accidental = "#"
        elif pitch_step == 1.5:
            accidental = "#+"
        else:
            accidental = str(pitch_step)

        return f"{self.accidental.get_pitch_parameters()[0]}{accidental}{self.octave}"

    # //public methods
    def add_child(self, child: [Accidental]) -> Accidental:
        """
        child's _update method will be called after adding.

        :param child:
        :return: child
        :rtype: :obj:`~musicscore.accidental.Accidental`
        """
        if self.parent_chord and self.parent_chord._finalized is True:
            raise AlreadyFinalizedError(self, "add_child")
        super().add_child(child)
        child._update()
        return child

    def add_tie(self, type: str) -> None:
        """
        A :obj:`~musicxml.xmlelement.xmlelement.XMLTie` child of type ``start`` or ``stop``  will be added to :obj:`~parent_note` object

        :param type: ``start``, ``stop``
        :exception: ValueError
        """
        if type not in ["start", "stop"]:
            raise ValueError
        self._ties.add(type)
        if self.parent_note:
            self.parent_note._update_ties()

    def get_pitch_or_rest(self) -> Union["XMLPitch", "XMLRest"]:
        """
        :return: :obj:`~musicxml.xmlelement.xmlelement.XMLPitch` or :obj:`~musicxml.xmlelement.xmlelement.XMLRest` object associated with this :obj:`~musicscore.midi.Midi`.
        """
        return self._pitch_or_rest

    def get_staff_number(self):
        """
        :return: get manually set staff number (necessary for cross-staff notation)
        """
        return self._staff_number

    def remove_tie(self, type: str) -> None:
        """
        :param type: ``start``, ``stop``
        :exception: ValueError
        """
        removed = False
        try:
            self._ties.remove(type)
            removed = True
        except KeyError:
            pass
        if removed and self.parent_note:
            self.parent_note._update_ties()

    def set_staff_number(self, val):
        """
        Set staff number manually (necessary for cross-staff notation).
        """
        self._staff_number = val

    def transpose(self, val: float) -> "Midi":
        """
        Adds val to value and returns self
        """
        self.value += val
        return self

    # //operators
    def __lt__(self, other):  # For x < y
        return self.value < other.value

    def __le__(self, other):  # For x <= y
        return self.value <= other.value

    def __gt__(self, other):  # For x > y
        return self.value > other.value

    def __ge__(self, other):  # For x >= y
        return self.value >= other.value

    def __copy__(self):
        copied = self.__class__(value=self.value, accidental=self.accidental)
        copied._ties = self._ties
        return copied

    def __deepcopy__(self, memodict={}):
        copied_accidental = copy.copy(self.accidental)
        copied = self.__class__(value=self.value, accidental=copied_accidental)
        copied._ties = copy.copy(self._ties)
        return copied

    def _copy_for_split(self):
        copied_accidental = copy.copy(self.accidental)
        copied = self.__class__(value=self.value, accidental=copied_accidental)
        copied.notehead = self.notehead
        return copied


class MidiNote(Midi):
    """
    Parent class of shorthand midi names: C, D, E, F, G, A, B.
    Example C(4, '#') = C(4, 'sharp') = C(4, sharp)
    """

    _VALUE = 60

    def __init__(
        self,
        octave: object,
        accidental_sign: object = None,
        *args: object,
        **kwargs: object,
    ) -> object:
        self._accidental_sign = accidental_sign
        self._octave = octave
        super().__init__(value=self._get_value(), *args, **kwargs)
        self._set_accidental_mode()

    def _set_accidental_mode(self):
        self.accidental.mode = get_accidental_mode(self.value, self._accidental_sign)

    def _get_value(self):
        return self._VALUE + self._get_accidental_value() - (4 - self._octave) * 12

    def _get_accidental_value(self):
        permitted = [
            "double-flat",
            "flat-flat",
            "ff",
            "bb",
            "three-quarters-flat",
            "flat",
            "f",
            "b",
            "quarter-flat",
            "quarter-sharp",
            "sharp",
            "s",
            "#",
            "three-quarters-sharp",
            "double-sharp",
            "sharp-sharp",
            "ss",
            "##",
            "x",
        ]
        if self._accidental_sign in ["double-flat", "flat-flat", "ff", "bb"]:
            return -2
        elif self._accidental_sign in ["three-quarters-flat"]:
            return -1.5
        elif self._accidental_sign in ["flat", "f", "b"]:
            return -1
        elif self._accidental_sign in ["quarter-flat"]:
            return -0.5
        elif self._accidental_sign in [None, "natural"]:
            return 0
        elif self._accidental_sign in ["quarter-sharp"]:
            return 0.5
        elif self._accidental_sign in ["sharp", "s", "#"]:
            return 1
        elif self._accidental_sign in ["three-quarters-sharp"]:
            return 1.5
        elif self._accidental_sign in ["double-sharp", "sharp-sharp", "ss", "##", "x"]:
            return 2
        else:
            raise ValueError(
                f"accidental_sign value {self._accidental_sign} can be None or must be in {permitted}"
            )

    def __repr__(self):
        return f"{self.name} at {id(self)}"

    def __copy__(self):
        copied = Midi(value=self.value, accidental=self.accidental)
        copied._ties = self._ties
        return copied

    def __deepcopy__(self, memodict={}):
        copied_accidental = copy.copy(self.accidental)
        copied = Midi(value=self.value, accidental=copied_accidental)
        copied._ties = copy.copy(self._ties)
        return copied

    def _copy_for_split(self):
        copied_accidental = copy.copy(self.accidental)
        copied = Midi(value=self.value, accidental=copied_accidental)
        copied.notehead = self.notehead
        return copied


class C(MidiNote):
    _VALUE = 60


class D(MidiNote):
    _VALUE = 62


class E(MidiNote):
    _VALUE = 64


class F(MidiNote):
    _VALUE = 65


class G(MidiNote):
    _VALUE = 67


class A(MidiNote):
    _VALUE = 69


class B(MidiNote):
    _VALUE = 71



"""
All <notation> elements defined in MusicXML. The lyrics show the notation assigned to each note.
"""

from pathlib import Path

from musicscore import Score, Chord, C, E, G
from unittest import TestCase
from musicscore.util import XML_DYNAMIC_CLASSES
from musicxml.xmlelement.xmlelement import (
    XMLFermata,
    XMLArpeggiate,
    XMLNonArpeggiate,
    XMLAccidentalMark,
    XMLAccent,
    XMLStrongAccent,
    XMLStaccato,
    XMLTenuto,
    XMLStaccatissimo,
    XMLSpiccato,
    XMLScoop,
    XMLPlop,
    XMLDoit,
    XMLFalloff,
    XMLBreathMark,
    XMLCaesura,
    XMLStress,
    XMLUnstress,
    XMLTrillMark,
    XMLTurn,
    XMLDelayedTurn,
    XMLInvertedTurn,
    XMLShake,
    XMLWavyLine,
    XMLDetachedLegato,
    XMLMordent,
    XMLInvertedMordent,
    XMLSchleifer,
    XMLTremolo,
    XMLUpBow,
    XMLDownBow,
    XMLHarmonic,
    XMLNatural,
    XMLArtificial,
    XMLBasePitch,
    XMLTouchingPitch,
    XMLSoundingPitch,
    XMLOpenString,
    XMLFingering,
    XMLThumbPosition,
    XMLPluck,
    XMLDoubleTongue,
    XMLTripleTongue,
    XMLStopped,
    XMLSnapPizzicato,
    XMLFret,
    XMLString,
    XMLPullOff,
    XMLHammerOn,
    XMLBend,
    XMLBendAlter,
    XMLWithBar,
    XMLPreBend,
    XMLRelease,
    XMLTap,
    XMLHeel,
    XMLToe,
    XMLFingernails,
    XMLF,
    XMLPpp,
    XMLSfp,
    XMLOtherDynamics,
)


class TestLily321(TestCase):
    def test_lily_321_Notations(self):
        score = Score()
        part = score.add_part("p1")

        xml_notations = [
            XMLFermata(),
            XMLFermata("normal"),
            XMLFermata("angled"),
            XMLFermata("square"),
            XMLFermata(type="inverted"),
            XMLArpeggiate(),
            XMLNonArpeggiate(),
            XMLAccidentalMark("double-sharp"),
            XMLAccent(),
            XMLStrongAccent(),
            XMLStaccato(),
            XMLTenuto(),
            XMLDetachedLegato(),
            XMLStaccatissimo(),
            XMLSpiccato(),
            XMLScoop(),
            XMLPlop(),
            XMLDoit(),
            XMLFalloff(),
            XMLBreathMark(),
            XMLCaesura(),
            XMLStress(),
            XMLUnstress(),
            None,
            XMLTrillMark(),
            XMLTurn(),
            XMLDelayedTurn(),
            XMLInvertedTurn(),
            XMLShake(),
            [
                XMLWavyLine(type="start"),
                XMLWavyLine(type="continue"),
                XMLWavyLine(type="stop"),
            ],
            XMLMordent(),
            XMLInvertedMordent(),
            XMLSchleifer(),
            XMLTremolo(3),
            (XMLTurn(), XMLAccidentalMark("natural", placement="below")),
            (
                XMLTurn(),
                XMLAccidentalMark("sharp"),
                XMLAccidentalMark("three-quarters-flat"),
            ),
            XMLUpBow(),
            XMLDownBow(),
            XMLHarmonic(),
        ]

        harmonics = [XMLHarmonic() for _ in range(6)]
        harmonics[0].add_child(XMLNatural())
        harmonics[2].add_child(XMLArtificial())
        harmonics[3].add_child(XMLNatural())
        harmonics[3].add_child(XMLBasePitch())
        harmonics[4].add_child(XMLNatural())
        harmonics[4].add_child(XMLTouchingPitch())
        harmonics[5].add_child(XMLNatural())
        harmonics[5].add_child(XMLSoundingPitch())

        xml_notations += harmonics

        xml_notations += [XMLOpenString(), XMLThumbPosition()]

        fingerings = [XMLFingering() for _ in range(7)]
        fingerings[1].value_ = "1"
        fingerings[2].value_ = "2"
        fingerings[3].value_ = "3"
        fingerings[4].value_ = "4"
        fingerings[5].value_ = "5"
        fingerings[5].value_ = "something"
        fingerings.append((XMLFingering("2"), XMLFingering("3"), XMLFingering("5")))

        xml_notations += fingerings

        xml_notations += [
            XMLPluck(),
            XMLPluck("a"),
            XMLThumbPosition(),
            XMLDoubleTongue(),
            XMLTripleTongue(),
            XMLStopped(),
            XMLSnapPizzicato(),
            XMLFret(0),
            XMLFret(1),
            XMLString(1),
            XMLString(5),
            [XMLHammerOn(type="start"), XMLHammerOn(type="stop")],
            [XMLPullOff(type="start"), XMLPullOff(type="stop")],
        ]

        bends = [XMLBend() for _ in range(4)]
        bends[0].add_child(XMLBendAlter(3))
        bends[1].add_child(XMLBendAlter(3))
        bends[1].add_child(XMLWithBar())
        bends[2].add_child(XMLBendAlter(0.5))
        bends[2].add_child(XMLPreBend())
        bends[3].add_child(XMLBendAlter(3.5))
        bends[3].add_child(XMLRelease())

        xml_notations += bends

        xml_notations += [
            XMLTap(),
            XMLTap(hand="left"),
            XMLHeel(),
            XMLToe(),
            XMLFingernails(),
            None,
            None,
            None,
            XMLF(),
            XMLPpp(),
            XMLSfp(),
            XMLOtherDynamics("sfffz"),
            (XMLStrongAccent(), XMLStaccato()),
        ]

        for n in xml_notations:
            if n is None:
                ch = Chord(0, 1)
                part.add_chord(ch)
            elif isinstance(n, XMLNonArpeggiate):
                ch = Chord([C(5), E(5), G(5)], 1)
                ch.add_lyric("XMLNonArpeggiate not implemented")
                part.add_chord(ch)
            elif isinstance(n, XMLArpeggiate):
                ch = Chord([C(5), E(5), G(5)], 1)
                ch.add_x(n)
                part.add_chord(ch)
            elif isinstance(n, XMLAccidentalMark):
                ch = Chord(C(5), 1)
                ch.add_x(n, parent_type="notation")
                part.add_chord(ch)
            elif isinstance(n, list):
                for x in n:
                    ch = Chord(C(5), 1)
                    ch.add_x(x)
                    part.add_chord(ch)
            elif isinstance(n, tuple):
                ch = Chord(C(5), 1)
                for x in n:
                    if isinstance(x, XMLAccidentalMark):
                        ch.add_x(x, parent_type="ornament")
                    else:
                        ch.add_x(x)
                part.add_chord(ch)
            elif n.__class__ in XML_DYNAMIC_CLASSES:
                ch = Chord(C(5), 1)
                ch.add_x(n, parent_type="notation", placement="below")
                part.add_chord(ch)
            else:
                ch = Chord(C(5), 1)
                ch.add_x(n)
                part.add_chord(ch)

        objects = (XMLStaccato(), XMLAccent(), XMLTenuto())
        placements = ["above", "below", "below"]
        ch = Chord(60, 1)
        for o, p in zip(objects, placements):
            ch.add_x(o, placement=p)
            # print(o.placement)
        part.add_chord(ch)

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



"""
Chords as grace notes.
"""

from pathlib import Path

from musicscore import Score, Chord, A, E, G

score = Score()
part = score.add_part("p1")

chords = [Chord(E(5), 2), Chord(E(5), 2)]
chords[1].add_grace_chord(G(5), type="16th", position="after")
chords[1].add_grace_chord(A(5), type="16th", position="after")

for ch in chords:
    part.add_chord(ch)

xml_path = Path(__file__).with_suffix(".xml")
score.export_xml(xml_path)



from fractions import Fraction
from unittest import TestCase

from musicscore import Time
from musicscore.chord import Chord
from musicscore.measure import Measure
from musicscore.tests.test_beat import create_voice
from musicscore.tests.util_subdivisions import (
    generate_all_quintuplets_manually,
    generate_all_sextuplets_manually,
    generate_all_triplets_manually,
)


class TestTuplets(TestCase):
    def test_simple_triplet(self):
        expected_1 = """<note>
    <pitch>
      <step>C</step>
      <octave>4</octave>
    </pitch>
    <duration>1</duration>
    <voice>1</voice>
    <type>eighth</type>
    <time-modification>
      <actual-notes>3</actual-notes>
      <normal-notes>2</normal-notes>
      <normal-type>eighth</normal-type>
    </time-modification>
    <notations>
      <tuplet type="start" number="1" bracket="yes" />
    </notations>
  </note>
"""
        expected_2 = """<note>
    <pitch>
      <step>C</step>
      <octave>4</octave>
    </pitch>
    <duration>2</duration>
    <voice>1</voice>
    <type>quarter</type>
    <time-modification>
      <actual-notes>3</actual-notes>
      <normal-notes>2</normal-notes>
      <normal-type>eighth</normal-type>
    </time-modification>
    <notations>
      <tuplet type="stop" number="1" />
    </notations>
  </note>
"""
        m = Measure(1)
        ch1 = Chord(midis=60, quarter_duration=1 / 3)
        ch2 = Chord(midis=60, quarter_duration=2 / 3)
        for c in [ch1, ch2]:
            c.midis[0].accidental.show = False
            m._add_chord(c)
        m.finalize()
        assert ch1.notes[0].to_string() == expected_1
        assert ch2.notes[0].to_string() == expected_2

    def test_chord_simple_triplet(self):
        m = Measure(1)
        chords = [Chord(60, 1 / 3), Chord(61, 2 / 3)]

        for x in chords:
            m._add_chord(x)

        m.finalize()

        t1, t2 = [ch.notes[0].xml_notations.xml_tuplet for ch in chords]
        assert t1.type == "start"
        assert t2.type == "stop"
        for note in [ch.notes[0] for ch in chords]:
            assert note.xml_time_modification is not None
            assert note.xml_time_modification.xml_actual_notes.value_ == 3
            assert note.xml_time_modification.xml_normal_notes.value_ == 2
            assert note.xml_time_modification.xml_normal_type.value_ == "eighth"

    def test_chord_quintuplet(self):
        measures = []
        for index, quintuplet in enumerate(generate_all_quintuplets_manually()):
            m = Measure(index + 1)
            for q in quintuplet:
                m._add_chord(Chord(midis=60, quarter_duration=q))
            m.finalize()
            measures.append(m)

        for m in measures:
            chords = [chord for chord in m.get_chords() if not chord.is_rest]
            for ch in chords:
                note = ch.notes[0]
                assert note.xml_time_modification is not None
                assert note.xml_time_modification.xml_actual_notes.value_ == 5
                assert note.xml_time_modification.xml_normal_notes.value_ == 4
                assert note.xml_time_modification.xml_normal_type.value_ == "16th"
            first_note = chords[0].notes[0]
            last_note = chords[-1].notes[0]
            t1, t2 = (
                first_note.xml_notations.xml_tuplet,
                last_note.xml_notations.xml_tuplet,
            )
            assert t1.type == "start"
            assert t2.type == "stop"

    def test_chord_sextuplets_not_writable_1(self):
        m1 = Measure(1)
        quarter_durations = [Fraction(1, 6), Fraction(5, 6)]
        for q in quarter_durations:
            m1._add_chord(Chord(midis=60, quarter_duration=q))
        m1._split_not_writable_chords()
        m1.finalize()
        b = m1.get_voice(staff_number=1, voice_number=1).get_children()[0]
        assert b.get_children() == m1.get_chords()[:3]
        n1, n2, n3 = [ch.notes[0] for ch in m1.get_chords()[:3]]
        # print([n.quarter_duration for n in [n1, n2, n3]])
        assert n1.xml_notations.xml_tuplet.type == "start"
        assert n3.xml_notations.xml_tuplet.type == "stop"
        assert not n1.is_tied
        assert n2.is_tied
        assert n3.is_tied_to_previous
        assert n3.xml_dot is not None
        assert n3.midi.accidental.show is False

    def test_chord_sextuplets_not_writable_2(self):
        m1 = Measure(1)
        quarter_durations = [Fraction(5, 6), Fraction(1, 6)]
        for q in quarter_durations:
            m1._add_chord(Chord(midis=60, quarter_duration=q))
        m1._split_not_writable_chords()
        m1.finalize()
        n1, n2, n3 = [ch.notes[0] for ch in m1.get_chords()[:3]]
        assert n1.xml_notations.xml_tuplet.type == "start"
        assert n3.xml_notations.xml_tuplet.type == "stop"
        assert n1.is_tied
        assert n2.is_tied_to_previous
        assert not n3.is_tied
        assert n1.xml_dot is not None
        assert n2.midi.accidental.show is False

    def test_chord_sextuplets(self):
        measures = []
        for index, sextuplet in enumerate(generate_all_sextuplets_manually()):
            m = Measure(index + 1)
            for q in sextuplet:
                m._add_chord(Chord(midis=60, quarter_duration=q))
            m._split_not_writable_chords()
            m.finalize()
            measures.append(m)

        for m in measures:
            chords = [chord for chord in m.get_chords() if not chord.is_rest]
            for ch in chords:
                note = ch.notes[0]
                assert note.xml_time_modification is not None
                assert note.xml_time_modification.xml_actual_notes.value_ == 6
                assert note.xml_time_modification.xml_normal_notes.value_ == 4
                assert note.xml_time_modification.xml_normal_type.value_ == "16th"
            first_note = chords[0].notes[0]
            last_note = chords[-1].notes[0]
            t1, t2 = (
                first_note.xml_notations.xml_tuplet,
                last_note.xml_notations.xml_tuplet,
            )
            assert t1.type == "start"
            assert t2.type == "stop"

    def test_update_xml_brackets(self):
        v1 = create_voice()
        beats = v1.update_beats(1, 1, 1, 1)
        ch1, ch2, ch3 = chords = [Chord(60, 2 / 5), Chord(61, 2 / 5), Chord(62, 1 / 5)]
        for ch in chords:
            beats[0].add_child(ch)
        v1.up.up._update_divisions()
        beats[0].finalize()
        assert ch1.notes[0].xml_notations.xml_tuplet.type == "start"
        assert ch1.notes[0].xml_notations.xml_tuplet.bracket == "yes"
        assert ch1.notes[0].xml_notations.xml_tuplet.number == 1
        assert ch2.notes[0].xml_notations is None
        assert ch3.notes[0].xml_notations.xml_tuplet.type == "stop"
        assert ch3.notes[0].xml_notations.xml_tuplet.number == 1

    def test_complex_tuplet(self):
        v1 = create_voice()
        beats = v1.update_beats(1)
        for quarter_duration in [1 / 6, 1 / 6, 1 / 6, 1 / 10, 3 / 10, 1 / 10]:
            v1._add_chord(Chord(60, quarter_duration))
        v1.up.up._update_divisions()
        beats[0].finalize()
        n1, n2, n3, n4, n5, n6 = [ch.notes[0] for ch in v1.get_chords()]
        for n in [n1, n2, n3]:
            assert (
                n.xml_time_modification.xml_actual_notes.value_,
                n.xml_time_modification.xml_normal_notes.value_,
                n.xml_time_modification.xml_normal_type.value_,
            ) == (3, 2, "16th")
        for n in [n4, n5, n6]:
            assert (
                n.xml_time_modification.xml_actual_notes.value_,
                n.xml_time_modification.xml_normal_notes.value_,
                n.xml_time_modification.xml_normal_type.value_,
            ) == (5, 4, "32nd")
        assert n1.xml_notations.xml_tuplet.type == "start"
        assert n2.xml_notations is None
        assert n3.xml_notations.xml_tuplet.type == "stop"
        assert n4.xml_notations.xml_tuplet.type == "start"
        assert n5.xml_notations is None
        assert n6.xml_notations.xml_tuplet.type == "stop"

    def test_group_beams_triplets(self):
        v1 = create_voice()
        beats = v1.update_beats(1, 1, 1)
        for quarter_duration in [
            q for group in generate_all_triplets_manually() for q in group
        ]:
            v1._add_chord(Chord(60, quarter_duration))
        v1.up.up._update_divisions()
        for index, beat in enumerate(beats):
            beat.finalize()
            if index == 0:
                for i, c in enumerate(beat.get_children()):
                    beams = c.notes[0].find_children("XMLBeam")
                    assert len(beams) == 1
                    assert beams[0].number == 1
                    assert (
                        beams[0].value_ == "begin"
                        if i == 0
                        else "continue"
                        if i == 1
                        else "end"
                    )
            else:
                for c in beat.get_children():
                    assert c.notes[0].find_child("XMLBeam") is None

    def test_quarter_triplets(self):
        m = Measure(1)
        m.time = Time(3, 2)
        quarter_durations = [2 / 3, 4 / 3, 4 / 3, 2 / 3, 2 / 3, 2 / 3, 2 / 3]
        chords = [Chord(60 + i, qd) for i, qd in enumerate(quarter_durations)]
        [m._add_chord(chord) for chord in chords]
        m.finalize()

        # check durations
        notes = [ch.notes[0] for ch in chords]
        divisions = m.xml_object.xml_attributes.xml_divisions.value_
        assert divisions == 3
        assert [n.xml_duration.value_ for n in notes] == [
            qd * divisions for qd in quarter_durations
        ]

        # check note types
        assert [n.xml_type.value_ for n in notes] == [
            "quarter",
            "half",
            "half",
            "quarter",
            "quarter",
            "quarter",
            "quarter",
        ]
        # check time modifications
        """
        <time-modification>
          <actual-notes>3</actual-notes>
          <normal-notes>2</normal-notes>
          <normal-type>quarter</normal-type>
        </time-modification>
        """
        for n in notes:
            assert (
                n.xml_time_modification.xml_actual_notes.value_,
                n.xml_time_modification.xml_normal_notes.value_,
                n.xml_time_modification.xml_normal_type.value_,
            ) == (3, 2, "quarter")

        # check brackets
        """
        <notations>
            <tuplet bracket="yes" number="1" type="start" />
        </notations>
        
        <notations>
            <tuplet number="1" type="stop" />
        </notations>
        """
        for i, n in enumerate(notes):
            if i in [0, 2, 4]:
                tuplet = n.xml_notations.xml_tuplet
                assert tuplet.bracket == "yes"
                assert tuplet.type == "start"
            elif i in [1, 3, 6]:
                tuplet = n.xml_notations.xml_tuplet
                assert tuplet.type == "stop"
            else:
                assert not n.xml_notations

    def test_quarter_triplets_with_actual_signatures(self):
        m = Measure(1)
        m.time = Time(6, 4)
        m.time.actual_signatures = [1, 2, 1, 2, 1, 2]
        quarter_durations = [2 / 3, 4 / 3, 4 / 3, 2 / 3, 2 / 3, 2 / 3, 2 / 3]
        chords = [Chord(60 + i, qd) for i, qd in enumerate(quarter_durations)]
        [m._add_chord(chord) for chord in chords]
        m.finalize()

        # check durations
        notes = [ch.notes[0] for ch in chords]
        divisions = m.xml_object.xml_attributes.xml_divisions.value_
        assert divisions == 3
        assert [n.xml_duration.value_ for n in notes] == [
            qd * divisions for qd in quarter_durations
        ]

        # check note types
        assert [n.xml_type.value_ for n in notes] == [
            "quarter",
            "half",
            "half",
            "quarter",
            "quarter",
            "quarter",
            "quarter",
        ]
        # check time modifications
        """
        <time-modification>
          <actual-notes>3</actual-notes>
          <normal-notes>2</normal-notes>
          <normal-type>quarter</normal-type>
        </time-modification>
        """
        for n in notes:
            assert (
                n.xml_time_modification.xml_actual_notes.value_,
                n.xml_time_modification.xml_normal_notes.value_,
                n.xml_time_modification.xml_normal_type.value_,
            ) == (3, 2, "quarter")

        # check brackets
        """
        <notations>
            <tuplet bracket="yes" number="1" type="start" />
        </notations>

        <notations>
            <tuplet number="1" type="stop" />
        </notations>
        """
        for i, n in enumerate(notes):
            if i in [0, 2, 4]:
                tuplet = n.xml_notations.xml_tuplet
                assert tuplet.bracket == "yes"
                assert tuplet.type == "start"
            elif i in [1, 3, 6]:
                tuplet = n.xml_notations.xml_tuplet
                assert tuplet.type == "stop"
            else:
                assert not n.xml_notations



from musicscore.exceptions import (
    AlreadyFinalizedError,
    ClassHasNoMusicXMLEquivalentError,
)


class FinalizeMixin:
    """
    FinalizeMixin is a mixin for finalizable classes: Score, Part, Measure, Staff, Voice, Beat and Chord
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._finalized = False

    def finalize(self) -> None:
        """
        :obj:`~musicscore.finalize.FinalizeMixin` method

        finalize can only be called once.

        It calls finalize method of all children.
        """
        if self._finalized:
            raise AlreadyFinalizedError(self)

        for child in self.get_children():
            if not child._finalized:
                child.finalize()

        self._finalized = True

    def to_string(self, *args, **kwargs):
        """
        :obj:`~musicscore.finalize.FinalizeMixin` method

        It triggers :obj:`~finalize()` first before calling parent's ``xml_object.to_string()`` method. If no xml_object exists (it means there is no direct MusicXML equavalent of this class) a
        :obj:`~musicscore.exceptions.ClassHasNoMusicXMLEquivalentError` is thrown.
        """

        if not self._finalized:
            self.finalize()

        try:
            return super().to_string(*args, **kwargs)
        except AttributeError:
            raise ClassHasNoMusicXMLEquivalentError(
                f"{self.__class__} has no direct equivalent in MusicXML and cannot be converted to string."
            )



from typing import List, Optional

from musicscore.exceptions import TimeActualSignaturesNotValidError
from musicscore.quarterduration import QuarterDuration
from musicscore.util import isinstance_as_string
from musicxml.xmlelement.xmlelement import XMLTime, XMLBeats, XMLBeatType
from musicscore.xmlwrapper import XMLWrapper

__all__ = ["Time", "flatten_times", "CONVERSION_DICTIONARY"]

#: If :obj:`Time.actual_signatures` is not set manually first this dictionary is used to create actual signature.
CONVERSION_DICTIONARY = {
    "2/8": [2, 8],
    "4/8": [2, 8, 2, 8],
    "5/8": [3, 8, 2, 8],
    "7/8": [4, 8, 3, 8],
}


def _convert_signatures_to_ints(signatures):
    output = []
    for i in range(int(len(signatures) / 2)):
        beat = signatures[i * 2]
        beat_type = signatures[i * 2 + 1]
        if isinstance(beat, str):
            l = beat.split("+")
            if len(l) == 1:
                output.append(int(beat))
                output.append(int(beat_type))
            else:
                for x in l:
                    output.append(int(x))
                    output.append(int(beat_type))
        else:
            output.append(beat)
            output.append(int(beat_type))
    if not output:
        raise TimeActualSignaturesNotValidError
    return output


def _get_quarter_durations_from_ints(signatures):
    output = 0
    for i in range(int(len(signatures) / 2)):
        beat = signatures[i * 2]
        beat_type = signatures[i * 2 + 1]
        output += 4 * beat / beat_type
    return output


class Time(XMLWrapper):
    _ATTRIBUTES = {"signatures", "actual_signatures", "parent_measure", "show"}
    XMLClass = XMLTime

    def __init__(self, *signatures, show=True, **kwargs):
        super().__init__()
        self._xml_object = self.XMLClass(**kwargs)
        self._parent_measure = None

        self._signatures = None
        self._actual_signatures = None
        self._intern_actual_signatures = None
        self._show = None

        self.signatures = signatures
        self.show = show

    def _calculate_actual_signatures(self):
        signatures = _convert_signatures_to_ints(self.signatures)
        signatures = [signatures[i : i + 2] for i in range(0, len(signatures), 2)]
        self._intern_actual_signatures = []
        for signature in signatures:
            key = "/".join([str(x) for x in signature])
            if key in CONVERSION_DICTIONARY:
                self._intern_actual_signatures.extend(CONVERSION_DICTIONARY[key])
            elif signature[1] % 8 == 0 and signature[0] % 3 == 0:
                self._intern_actual_signatures.extend(
                    [3, signature[1]] * (signature[0] // 3)
                )
            else:
                self._intern_actual_signatures.extend([1, signature[1]] * signature[0])
        return self._intern_actual_signatures

    def _reset_actual_signatures(self) -> None:
        """
        Resets actual signatures to None.
        """
        self._actual_signatures = None
        self._intern_actual_signatures = None

    def _update_signature_objects(self):
        signatures = [
            self.signatures[i : i + 2] for i in range(0, len(self.signatures), 2)
        ]
        for beats, beat_type in zip(
            self._xml_object.find_children("XMLBeats"),
            self._xml_object.find_children("XMLBeatType"),
        ):
            if signatures:
                signature = signatures.pop(0)
                beats.value_ = str(signature[0])
                beat_type.value_ = str(signature[1])
            else:
                beats.up.remove(beats)
                beat_type.up.remove(beat_type)
        for beats, beat_type in signatures:
            self._xml_object.add_child(XMLBeats(str(beats)))
            self._xml_object.add_child(XMLBeatType(str(beat_type)))

    @property
    def actual_signatures(self) -> List[int]:
        """
        Set and gets actual signatures. If :obj:`parent_measure` exists its beats inside voices will be updated.

        :return: A list of int representing actual time signatures. If not set manually, it is calculated internally. For example a 4/4 time
                 signature gets automatically [1, 4, 1, 4, 1, 4, 1, 4] as actual_signatures if not set otherwise.

        .. seealso::
           :obj:`CONVERSION_DICTIONARY`
        """
        if self._actual_signatures is None:
            if self._intern_actual_signatures is None:
                self._calculate_actual_signatures()
            return self._intern_actual_signatures
        return self._actual_signatures

    @actual_signatures.setter
    def actual_signatures(self, val):
        if val is not None:
            val = _convert_signatures_to_ints(val)
        self._actual_signatures = val
        if self.parent_measure:
            self.parent_measure._update_voice_beats()

    @property
    def parent_measure(self) -> Optional["Measure"]:
        """
        Set and get parent :obj:`~musicscore.measure.Measure`.
        """
        return self._parent_measure

    @parent_measure.setter
    def parent_measure(self, val):
        if not val and not isinstance_as_string(val, "Measure"):
            raise TypeError
        self._parent_measure = val

    @property
    def signatures(self) -> List[int]:
        """
        Set and gets signatures. If :obj:`~parent_measure` exists, beats inside its voices will be updated. If it is set to ``None`` a 4/4 signature will be used.

        :return: A list of int representing time signature.
        """
        return self._signatures

    @signatures.setter
    def signatures(self, val):
        if not val:
            val = [4, 4]
        for v in val:
            if not isinstance(v, int) and not isinstance(v, str):
                raise TypeError
        self._signatures = val
        self._reset_actual_signatures()
        self._update_signature_objects()
        if self.parent_measure:
            self.parent_measure._update_voice_beats()

    @property
    def show(self) -> bool:
        """
        If time signature is shown or not.

        :type: bool
        :return: bool
        """
        return self._show

    @show.setter
    def show(self, val):
        if not isinstance(val, bool):
            raise TypeError
        self._show = val

    def get_beats_quarter_durations(self) -> List[QuarterDuration]:
        """
        :return: List of quarter durations according to :obj:`actual_signatures`
        """
        return [
            QuarterDuration(numerator, denominator) * 4
            for numerator, denominator in [
                self.actual_signatures[i : i + 2]
                for i in range(0, len(self.actual_signatures), 2)
            ]
        ]

    def __copy__(self):
        cp = self.__class__(*self.signatures, show=self.show)
        cp._actual_signatures = self._actual_signatures
        return cp

    def __rmul__(self, other):
        return [self.__copy__() for _ in range(other)]


def flatten_times(times) -> List[Time]:
    """
    :param times: an expandable list of times or tuples representing times. For example [x * Time(3, 8)] return x time intances of Time(3, 8)
    :return: List[Time]

    >>> ts = [2 * Time(3, 8), (3, 4), 3 * [(1, 8)], Time(1, 8, 3, 4), Time(3, 4)]
    >>> [t.signatures for t in flatten_times(ts)]
    [(3, 8), (3, 8), (3, 4), (1, 8), (1, 8), (1, 8), (1, 8, 3, 4), (3, 4)]
    """
    output = []
    for time in times:
        if isinstance(time, Time):
            output.append(time)
        elif hasattr(time, "__iter__"):
            if {isinstance(t, int) for t in time} == {True}:
                output.append(Time(*time))
            else:
                for t in time:
                    if isinstance(t, Time):
                        output.append(t)
                    else:
                        output.append(Time(*t))
    return output



from pathlib import Path

from fractions import Fraction
from unittest import TestCase

from musicscore.chord import Chord
from musicscore.part import Part
from musicscore.score import Score
from musicscore.tests.util import create_test_xml_paths
from musicscore.tests.util_subdivisions import generate_all_subdivision_patterns

path = Path(__file__)


class TestTuplets1(TestCase):
    def test_tuplets_1(self):
        """
        Write all possible tuplet combinations up until septuplets
        """

        s = Score()

        p = s.add_child(Part("P1", name="Music"))

        p.add_measure(time=(1, 4))

        rhythmic_patterns = []
        for subdivision in range(1, 8):
            rhythmic_patterns.extend(
                generate_all_subdivision_patterns(subdivision, True)
            )

        for rhythmic_pattern in rhythmic_patterns:
            subdivision = sum(rhythmic_pattern)
            for x in rhythmic_pattern:
                ch = Chord(60, Fraction(x, subdivision))
                ch.add_lyric(x)
                p.add_chord(ch)
        xml_path = create_test_xml_paths(path, "upto_7")[0]
        s.export_xml(xml_path)

    def test_tuplets_8(self):
        """
        Write all possible tuplet combinations of 32nd
        """

        s = Score()

        p = s.add_child(Part("P1", name="Music"))

        p.add_measure(time=(1, 4))

        subdivision = 8
        rhythmic_patterns = generate_all_subdivision_patterns(subdivision, True)
        for x in [d for rp in rhythmic_patterns for d in rp]:
            ch = Chord(60, Fraction(x, subdivision))
            ch.add_lyric(x)
            p.add_chord(ch)
        xml_path = create_test_xml_paths(path, "8")[0]
        s.export_xml(xml_path)

    def test_tuplets_9(self):
        """
        Write all possible tuplet combinations of nonuplets
        """

        s = Score()

        p = s.add_child(Part("P1", name="Music"))

        p.add_measure(time=(1, 4))

        subdivision = 9
        rhythmic_patterns = generate_all_subdivision_patterns(subdivision, True)
        for x in [d for rp in rhythmic_patterns for d in rp]:
            ch = Chord(60, Fraction(x, subdivision))
            ch.add_lyric(x)
            p.add_chord(ch)
        xml_path = create_test_xml_paths(path, "9")[0]
        s.export_xml(xml_path)



from musicscore import Chord, Part
from unittest import TestCase
from musicxml import XMLDot


class TestWritingBeats(TestCase):
    def test_writing_32nds(self):
        # 1, 5, 2
        values = [1, 5, 2]
        unit = 8
        chords = [Chord(60, v / unit) for v in values]
        p = Part(id="p1")
        m = p.add_measure([1, 4])
        [p.add_chord(ch) for ch in chords]
        beat = m.get_beats()[0]
        assert beat.get_chords() == chords
        assert chords[1].quarter_duration == 5 / 8
        m.finalize()
        """After finalizing the qd of the second chord has been changed:"""
        assert chords[1].quarter_duration == 3 / 8
        """This is because 5/8 qd is not writable and has been split in two qds: 3/8 + 2/8.  It means a chord has
        been added to the beat after the second chord qd of which has been changed too."""
        assert [ch.quarter_duration for ch in beat.get_chords()] == [
            1 / 8,
            3 / 8,
            2 / 8,
            2 / 8,
        ]
        """Be aware now that chords and beat.get_chords() diverged"""
        assert beat.get_chords() != chords
        """and the second chord is not copied but has destructively changed its qd:"""
        assert [ch.quarter_duration for ch in chords] == [1 / 8, 3 / 8, 2 / 8]
        assert sum([ch.quarter_duration for ch in chords]) != 1
        """Of course there the spit chords are tied:"""
        assert beat.get_chords()[1].is_tied_to_next
        assert beat.get_chords()[2].is_tied_to_previous
        """Which quarter durations (at which position or offset in Beat) are not writable or are wished otherwise to
        be split are recorded in a dictionary named SPLITTABLES inside config.py file."""

        """The type of each chord is (at the moment) set on its Note children via its QuarterDuration. Also for this 
        purpose there is a dictionary inside config.py named NOTTYPES"""
        assert [ch.quarter_duration.get_type() for ch in beat.get_chords()] == [
            "32nd",
            "16th",
            "16th",
            "16th",
        ]
        assert [ch.get_children()[0].xml_type.value_ for ch in beat.get_chords()] == [
            "32nd",
            "16th",
            "16th",
            "16th",
        ]
        """
        todo:
        It would be better to add a property to Chord and Note which can be set also manually."""

        """QuarterDuration._get_type_and_dots() returns a tuple (type, number_of_dots). Number of dots is tried to be 
        determined automatically. If this attempt is not successful an Exception is raised.
        Note.update_dots() is used to set note's number of dots
        """
        assert [
            ch.quarter_duration.get_number_of_dots() for ch in beat.get_chords()
        ] == [0, 1, 0, 0]
        assert [
            len(ch.get_children()[0].xml_object.get_children_of_type(XMLDot))
            for ch in beat.get_chords()
        ] == [0, 1, 0, 0]
        """beams: Beat.beam_chord_group() Too complicated. Can it be simplified? """



"""
A huge orchestra score with 28 parts and different kinds of nested bracketed groups. Each part/group is assigned a
name and an abbreviation to be shown before the staff. Also, most of the groups show unbroken barlines,
while the barlines are broken between the groups.
"""

from pathlib import Path

from musicscore import Score, Chord, TrebleClef, BassClef
from musicscore.clef import PercussionClef, AltoClef
from unittest import TestCase
from musicxml import XMLTranspose


class TestLily41c(TestCase):
    def test_lily_41c_StaffGroups(self):
        score = Score()
        score.page_layout.size = "A4"
        score.scaling.millimeters = 2.6176
        score.scaling.tenths = 40
        score.page_layout.margins.left = 200
        instruments = {
            "Piccolo": {"abbreviation": "picc", "clef": TrebleClef(octave_change=1)},
            "Flute 1": {"abbreviation": "fl 1"},
            "Flute 2": {"abbreviation": "fl 2"},
            "Oboe 1": {"abbreviation": "ob 1"},
            "Oboe 2": {"abbreviation": "ob 2"},
            "English Horn": {"abbreviation": "eh", "transposition": -7},
            "Clarinet in Eb": {"abbreviation": "cl(Eb)", "transposition": 3},
            "Clarinet 1 in Bb": {"abbreviation": "cl(Bb) 1", "transposition": -2},
            "Clarinet 2 in Bb": {"abbreviation": "cl(Bb) 2", "transposition": -2},
            "Bass Clarinet": {
                "abbreviation": "bcl",
                "clef": BassClef(octave_change=-1),
                "transpose": -2,
            },
            "Bassoon 1": {"abbreviation": "bn 2", "clef": BassClef()},
            "Bassoon 2": {"abbreviation": "bn 1", "clef": BassClef()},
            "Contrabassoon": {
                "abbreviation": "cbn",
                "clef": BassClef(octave_change=-1),
            },
            "Horn in F 1": {"abbreviation": "hn(F) 1", "transposition": -7},
            "Horn in F 2": {"abbreviation": "hn(F) 2", "transposition": -7},
            "Trumpet in C 1": {"abbreviation": "tp(C) 1"},
            "Trumpet in C 2": {
                "abbreviation": "tp(C) 1",
            },
            "Trombone 1": {"abbreviation": "tbn", "clef": BassClef()},
            "Trombone 2": {"abbreviation": "tbn", "clef": BassClef()},
            "Tuba": {"abbreviation": "tb", "clef": BassClef()},
            "Timpani": {"abbreviation": "tmp", "clef": BassClef()},
            "Percussion": {"abbreviation": "perc", "clef": PercussionClef()},
            "Harp": {"abbreviation": "hrp", "number_of_staves": 2},
            "Piano": {"abbreviation": "pno", "number_of_staves": 2},
            "Violin I": {
                "abbreviation": "vn 2",
            },
            "Violin II": {
                "abbreviation": "vn 1",
            },
            "Viola": {"abbreviation": "va", "clef": AltoClef()},
            "Cello": {"abbreviation": "vc", "clef": BassClef()},
            "Contrabass": {"abbreviation": "cb", "clef": BassClef(octave_change=-1)},
        }
        for i, k in enumerate(instruments.keys()):
            part = score.add_part(f"p{i + 1}")
            part.name = k
            part.abbreviation = instruments[k]["abbreviation"]

            part.add_chord(Chord(61, 4))
            clef = instruments[k].get("clef")

            if clef:
                part.get_current_measure().get_staff(1).clef = clef
            number_of_staves = instruments[k].get("number_of_staves")
            if number_of_staves == 2:
                part.add_chord(Chord(0, 4), staff_number=2)
            transposition = instruments[k].get("transposition")
            if transposition:
                tr = XMLTranspose()
                tr.xml_chromatic = transposition
                part.get_current_measure().xml_attributes.xml_transpose = tr
            part.add_chord(Chord(0, 4))
            if number_of_staves == 2:
                part.add_chord(Chord(0, 4), staff_number=2)
            part.get_current_measure().new_system = True
        score.group_parts(2, 1, 13, "bracket")
        score.group_parts(2, 14, 20, "bracket")
        score.group_parts(2, 25, 29, "bracket")

        score.group_parts(1, 2, 3, "square", default_x=-18)
        score.group_parts(1, 4, 5, "square", default_x=-18)
        score.group_parts(1, 8, 9, "square", default_x=-18)
        score.group_parts(1, 11, 12, "square", default_x=-18)
        score.group_parts(1, 14, 15, "square", default_x=-18)
        score.group_parts(1, 16, 17, "square", default_x=-18)
        score.group_parts(1, 18, 19, "square", default_x=-18)
        score.group_parts(1, 25, 26, "square", default_x=-18)

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



"""
Chords as grace notes.
"""

from pathlib import Path

from musicscore import Score, Chord, C, A, D, F, B
from unittest import TestCase
from musicxml.xmlelement.xmlelement import XMLSlur


class TestLily24b(TestCase):
    def test_lily_24b_ChordAsGraceNote(self):
        score = Score()
        part = score.add_part("p1")

        chords = [Chord(C(5), 1), Chord(C(5), 1), Chord([C(5), A(4)], 1)]
        gch = chords[1].add_grace_chord([D(5), F(5)])
        gch.add_x(XMLSlur(type="start"))
        chords[1].add_x(XMLSlur(type="stop"))

        gch = chords[2].add_grace_chord([B(4), D(5)])
        gch.add_x(XMLSlur(type="start", orientation="under"))
        chords[2].add_x(XMLSlur(type="stop"))

        for ch in chords:
            part.add_chord(ch)

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



"""
Part names and abbreviations can contain line breaks.
"""

from pathlib import Path

from musicscore import Score, Chord, B
from unittest import TestCase


class TestLily413(TestCase):
    def test_lily_41e_StaffGroups_InstrumentNames_Linebroken(self):
        score = Score()
        part = score.add_part("p1")
        part.name = """Long
Staff
Name        
"""
        part.abbreviation = """St.
Nm.
"""
        [part.add_chord(Chord(B(4), 4)) for _ in range(23)]
        part.get_measure(6).new_system = True
        part.get_measure(15).new_system = True
        for m in part.get_children()[13:]:
            m.width = 80

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



from unittest import TestCase
from musicxml import XMLBarline
from pathlib import Path

from musicscore import Score, Chord

"""
Barlines can appear at mid-measure positions, without using an implicit measure!
"""

"""
MuseScore supports mid-measure barlines
"""


class TestLily46b(TestCase):
    def test_lily_46b_MidmeasureBarlines(self):
        score = Score()
        p = score.add_part("p1")
        midis = [72, 69, 65, 72]
        chords = [Chord(m, 1) for m in midis]
        [p.add_chord(ch) for ch in chords]
        b = XMLBarline(location="middle")
        b.xml_bar_style = "dashed"
        chords[1].add_xml_element_after_notes(b)
        score.finalize()

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)
